#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo, R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

class ACBN0:

  def __init__ ( self, prefix, pthr=0.95, workdir='./', mpi_qe='', nproc=1, qe_path='', qe_options='', mpi_python='', python_path='' ):
    from .defs.file_io import struct_from_inputfile_QE
    from .defs.upf_gaussfit import gaussian_fit
    from .defs.header import header
    from os.path import join
    from os import chdir

    header()
    print('\nPerforming ACBN0 self-consistent determination of Hubbard U corrections.\n')

    self.prefix = prefix
    self.pthr = pthr
    self.workdir = workdir
    self.mpi_qe = mpi_qe
    self.nproc = nproc
    self.qpath = qe_path
    self.mpi_python = mpi_python
    self.ppath = python_path
    self.qoption = qe_options

    self.uVals = {}
    self.occ_states = {}
    self.occ_values = {}
    self.hubbard_occ = {}
    self.hubbard_tag = 'HUBBARD (ortho-atomic)'

    chdir(self.workdir)

    # Get structure information
    self.blocks,self.cards = struct_from_inputfile_QE(f'{self.prefix}.scf.in')
    self.nspin = int(struct['nspin']) if 'nspin' in self.blocks['system'] else 1

    # Generate gaussian fits
    print('Generating gaussian fits for pseudopotential basis states.\n')
    self.basis = {}
    self.uspecies = []
    for s in self.cards['ATOMIC_SPECIES'][1:]:
      ele,_,pp = s.split()
      self.uspecies.append(ele)
      atno,basis = gaussian_fit(pp)
      self.basis[ele] = basis

    # Store U values from input template
    if 'HUBBARD' in self.cards:
      import re

      self.hubbard_tag = self.cards['HUBBARD'][0]
      for h in self.cards['HUBBARD'][1:]:

        _,sym,uval = h.split()
        self.uVals[sym] = float(uval)

        ele,occ = sym.split('-')
        if ele not in self.occ_states:
          self.occ_states[ele] = []
        self.occ_states[ele].append(occ)

      # Store occupations from input template
      nat = len(self.uspecies)
      for s in self.blocks['system']:
        if 'hubbard_occ' in s:

          i,j = map(int, re.findall('\(([^\)]+),([^\)]+)\)', s)[0])
          if i > nat:
            msg = f'hubbard_occ index 1 (value:{i}) out of range for {nat} species.'
            raise ValueError(msg)

          spec = self.uspecies[i-1]
          nstates = len(self.occ_states[spec])
          if j > nstates:
            msg = f'hubbard_occ index 2 (value:{j}) out of range, {nstates} states listed for {spec}.'
            raise ValueError(msg)

          state = self.occ_states[spec][j-1]
          self.occ_values[f'{spec}-{state}'] = float(self.blocks['system'][s])
          self.hubbard_occ[s] = self.blocks['system'][s]


  def set_hubbard_parameters ( self, hubbard ):

    htype = type(hubbard)
    if htype in [list, tuple]:

      for h in hubbard:
        ele,occ = h.split('-')
        self.uVals[h] = 0.01
        if ele not in self.occ_states:
          self.occ_states[ele] = []
        self.occ_states[ele].append(occ)

    elif htype is dict:

      for k,v in hubbard.items():

        self.uVals[k] = 0.01
        if v is None:
          continue

        try:

          vtype = type(v)
          if vtype not in [list, tuple]:
            self.uVals[k] = float(v)

          else:
            if len(v) >= 1 and v[0] is not None:
              self.uVals[k] = v[0]

            if len(v) >= 2 and v[1] is not None:

              ele,occ = k.split('-')
              if ele not in self.occ_states:
                self.occ_states[ele] = [occ]

              elif occ not in self.occ_states[ele]:
                self.occ_states[ele].append(occ)

              i = 1 + self.uspecies.index(ele)
              j = 1 + self.occ_states[ele].index(occ)

              key = f'hubbard_occ({i},{j})'
              self.hubbard_occ[key] = float(v[1])

        except Exception as e:
          print('Dictionary values should either be the initial U value, or a list/tuple containing (initial U, hubbard occupation).')
          raise e

    else:
      msg = 'Input type should be either list or dict.'
      raise TypeError(msg)


  def optimize_hubbard_U ( self, convergence_threshold=0.01 ):

    print('\nBeginning self-consistent loop.\n')
    itr = 0
    converged = False
    while not converged:

      itr += 1
      print(f'Iteration #{itr}\n')

      self.run_dft(self.prefix, self.uspecies, self.uVals)

      save_prefix = self.blocks['control']['prefix'].strip('"').strip("'")
      self.run_paoflow(self.prefix, save_prefix)

      new_U = self.run_acbn0(self.prefix)

      converged = True
      print('\nNew U values:')
      for k,v in new_U.items():
        print(f'  {k} : {v}')
        if converged and np.abs(self.uVals[k]-v) > convergence_threshold:
          converged = False
      print('', flush=True)

      self.uVals = new_U


  def acbn0 ( self, fname ):

    data = {}
    with open(fname, 'r') as f:
      for l in f.readlines():
        if l.startswith('#'):
          continue
        k,v = (v.strip() for v in l.split('='))
        data[k] = v

    ang_to_bohr = 1.88973
    sym = data['symbol']
    nspin = int(data['nspin'])
    species = data['atlabels'].split(',')

    lattice = np.array([float(v) for v in data['latvects'].split(',')])
    lattice = lattice.reshape((3,3)) / ang_to_bohr

    coords = np.array([float(v) for v in data['coords'].split(',')])
    coords = coords.reshape((coords.shape[0]//3,3)) / ang_to_bohr

    split_bstr = lambda bstr : np.array([int(v) for v in bstr.split(',')])
    basis_dm = split_bstr(data['reduced_basis_dm'])
    basis_2e = split_bstr(data['reduced_basis_2e'])

    gauss_basis = self.getbasis(self.basis, species, lattice, coords)
    nbasis = len(gauss_basis)

    fpath = './output'
    kpnts,kwght,Sks,Hks_up,Hks_dw = self.read_ham_data(fpath, nspin)

    dk,nlm = self.Dk(basis_dm, basis_2e, Hks_up, Sks)
    nlm = self.Nmm(nlm, Hks_up, kwght)
    nnlm = nlm.shape[0]

    dk_dn = None
    den_U,den_J = 0,0
    if nspin == 1:
      Naa,Nab = 0., 0.
      for i1,m1 in enumerate(nlm):
        for i2,m2 in enumerate(nlm):
          nlm12 = m1 * m2
          Nab += nlm12
          if i1 != i2:
            Naa += nlm12
      den_U = 2 * (Naa.real + Nab.real)
      den_J = 2 * Naa.real

    else:
      dk_dn,nlmd = self.Dk(basis_dm, basis_2e, Hks_dw, Sks)
      nlmd = self.Nmm(nlmd, Hks_dw, kwght)

      Naa,Nbb,Nab = 0., 0., 0.
      for i1 in range(nnlm):
        for i2 in range(nnlm):
          Nab += nlm[i1] * nlmd[i2]
          if i1 != i2:
            Naa += nlm[i1] * nlm[i2]
            Nbb += nlmd[i1] * nlmd[i2]
      den_U = Naa.real + Nbb.real + 2*Nab.real
      den_J = Naa.real + Nbb.real

    DR_0 = self.DR(dk, kwght)

    DR_0_dn = DR_0
    if nspin == 2:
      DR_0_dn = self.DR(dk_dn, kwght)

    num_U,num_J = self.hartree_energy(DR_0, DR_0_dn, gauss_basis, basis_2e)

    hartree_to_eV = 27.211396132
    U = U_eff = hartree_to_eV * num_U / den_U
    if den_J == 0:
      J = 'Inf'
    else:
      J = hartree_to_eV * num_J / den_J
      U_eff -= J

    with open(f'{sym}_UJ.txt', 'w') as f:
      f.write(f'U : {U}\nJ : {J}\nU_eff : {U_eff}\n')


  def exec_command ( self, command ):
    from os import system

    print(f'Starting Process: {command}', flush=True)
    return system(command)


  def exec_QE ( self, executable, fname ):
    from os.path import join

    exe = join(self.qpath, executable)
    fout = fname.replace('in', 'out')
    
    command = f'{self.mpi_qe} {exe} {self.qoption} < {fname} > {fout}'
    return self.exec_command(command)


  def exec_PAOFLOW ( self ):
    from os.path import join

    prefix = join(self.ppath, 'python')
    command = f'{self.mpi_python} {prefix} acbn0.py > paoflow.out'
    return self.exec_command(command)


  def hubbard_card ( self ):

    if len(self.uVals) == 0:
      msg = 'No U found. Add U to the template inputfiles or with set_hubbard_parameters.'
      raise ValueError(msg)

    card = [self.hubbard_tag]
    for k,v in self.uVals.items():
      card.append(' U {} {}'.format(k, v))

    return card


  def run_dft ( self, prefix, species, uVals ):
    from .defs.file_io import create_atomic_inputfile
    from .defs.file_io import struct_from_inputfile_QE

    blocks,cards = struct_from_inputfile_QE(f'{prefix}.scf.in')
    cards['HUBBARD'] = self.hubbard_card()
    for k,v in self.hubbard_occ.items():
      blocks['system'][k] = v
    create_atomic_inputfile('scf', blocks, cards)

    blocks,cards = struct_from_inputfile_QE(f'{prefix}.nscf.in')
    cards['HUBBARD'] = self.hubbard_card()
    for k,v in self.hubbard_occ.items():
      blocks['system'][k] = v
    create_atomic_inputfile('nscf', blocks, cards)

    blocks,cards = struct_from_inputfile_QE(f'{prefix}.projwfc.in')
    create_atomic_inputfile('projwfc', blocks, cards)

    executables = {'scf':'pw.x', 'nscf':'pw.x', 'projwfc':'projwfc.x'}
    for c in ['scf', 'nscf', 'projwfc']:
      ecode = self.exec_QE(executables[c], f'{c}.in')


  def run_paoflow ( self, prefix, save_prefix ):
    from .defs.file_io import create_acbn0_inputfile
    from os import system

    fstr = f'{prefix}_PAO_bands' + '{}.in'
    calcs = []
    if self.nspin == 1:
      calcs.append(fstr.format('')) 
    else:
      calcs.append(fstr.format('_up'))
      calcs.append(fstr.format('_down'))

    create_acbn0_inputfile(save_prefix, self.pthr)
    ecode = self.exec_PAOFLOW()


  def read_cell_atoms ( self, fname ):

    lines = None
    with open(fname, 'r') as f:
      lines = f.readlines()

    il = 0
    while 'lattice parameter' not in lines[il]:
      il += 1
    alat = float(lines[il].split()[4])

    while 'number of atoms/cell' not in lines[il]:
      il += 1
    nat = int(lines[il].split()[4])

    while 'crystal axes:' not in lines[il]:
      il += 1
    il += 1
    lattice = np.array([[float(v) for v in lines[il+i].split()[3:6]] for i in range(3)])

    while 'site n.' not in lines[il]:
      il += 1
    il += 1
    species = []
    positions = np.empty((nat,3), dtype=float)
    for i in range(nat):
      ls = lines[il+i].split()
      species.append(ls[1])
      positions[i,:] = np.array([float(v) for v in ls[6:9]])

    lattice *= alat
    positions *= alat
    return lattice,species,positions


  def hubbard_orbital ( self, ele ):

    orb = ele[-1]
    orbitals = {'s':0, 'p':1, 'd':2}

    if orb in orbitals:
      return orbitals[orb]

    else:
      raise Exception(f'Element {ele} has no defined Hubbard orbital')


  def run_acbn0 ( self, prefix ):
    import re

    lattice,species,positions = self.read_cell_atoms('scf.out')
    lattice_string = ','.join([str(v) for a in lattice for v in a])
    position_string = ','.join([str(v) for a in positions for v in a])
    species_string = ','.join(species)

    sind = 0
    state_lines = open('projwfc.out', 'r').readlines()
    while 'state #' not in state_lines[sind]:
      sind += 1
    send = sind
    while 'state #' in state_lines[send]:
      send += 1
    state_lines = state_lines[sind:send]

    uVals = {}
    #for s in list(set(species)):
    for k,v in self.uVals.items():

      ostates = []
      ustates = []
      s = k.split('-')[0]
      horb = self.hubbard_orbital(k)
      for n,sl in enumerate(state_lines):
        stateN = re.findall('\(([^\)]+)\)', sl)
        oele = stateN[0].strip()
        oL = int(re.split('=| ', stateN[1])[1])
        if s in oele and oL == horb:
          ostates.append(n)
          if s == oele:
            ustates.append(n)
      sstates = [ustates[0]]
      for i,us in enumerate(ustates[1:]):
        if us == 1 + sstates[i]:
          sstates.append(us)
        else:
          break

      basis_dm = ','.join(list(map(str,ostates)))
      basis_2e = ','.join(list(map(str,sstates)))

      fname = f'{prefix}_acbn0_infile_{k}.txt'
      with open(fname, 'w') as f:
        nspin = self.nspin
        f.write(f'symbol = {k}\n')
        f.write(f'latvects = {lattice_string}\n')
        f.write(f'coords = {position_string}\n')
        f.write(f'atlabels = {species_string}\n')
        f.write(f'nspin = {nspin}\n')
        f.write(f'fpath = ./\n')
        f.write(f'outfile = {prefix}_acbn0_outfile_{k}.txt\n')
        f.write(f'reduced_basis_dm = {basis_dm}\n')
        f.write(f'reduced_basis_2e = {basis_2e}\n')

      self.acbn0(fname)
      with open(f'{k}_UJ.txt', 'r') as f:
        lines = f.readlines()
        uVals[k] = float(lines[2].split(':')[1])

    return uVals


  def getbasis ( self, basis, species, lattice, coords ):
    from .defs.pyints import CGBF
 
    basis_functions = []
    for a,pos in zip(species, coords):
      for shell in basis[a]:
        for subshell in shell:

          bf = CGBF(pos*1.88973, a)
          for lx,ly,lz,coeff,zeta in subshell:
            bf.pnorms.append(1.)
            bf.pexps.append(zeta)
            bf.pcoefs.append(coeff)
            bf.powers.append((lx,ly,lz))

          basis_functions.append(bf)

    return basis_functions


  def Dk ( self, basis_dm, basis_2e, Hks, Sks ):
    from scipy.linalg import eigh

    nbasis,_,nkpnts = Hks.shape
    size_dm,size_2e = basis_dm.shape[0],basis_2e.shape[0]
    D_k = np.zeros((nbasis,nbasis,nkpnts), dtype=complex)
    nlm_k = np.zeros((size_2e,nbasis,nkpnts), dtype=complex)

    # Find the density matrix for each k
    for ik in range(nkpnts):

      eig,vec = eigh(Hks[:,:,ik], Sks[:,:,ik])

      occ_ind = np.where(eig <= 0.)[0]
      nocc = len(occ_ind)

      lm_dm = np.zeros((size_dm,nocc), dtype=complex)
      lm_2e = np.zeros((size_2e,nocc), dtype=complex)

      sk_dm = Sks[basis_dm, :, ik]
      sk_2e = Sks[basis_2e, :, ik]
      for im in range(nocc):
        vim = vec[:, im]
        lm_dm[:,im] = np.conj(vim[basis_dm]) * sk_dm.dot(vim)
        lm_2e[:,im] = np.conj(vim[basis_2e]) * sk_2e.dot(vim)

      evec2 = vec[:, occ_ind]
      nlm_k[:,:nocc,ik] = lm_2e
      lm_dm = np.sum(lm_dm, axis=0)

      D_k[:,:,ik] = np.tensordot(np.conj(evec2*lm_dm), evec2, axes=([1],[1]))

    nlm_k[:,:,0] = 0
    return D_k, nlm_k


  def Nmm ( self, nlm, Hks, kwght ):

    lm_size,nbasis,nkp = nlm.shape
    nlm_aux = np.zeros((lm_size,nbasis), dtype=complex)
    for ik,wght in enumerate(kwght):
      nlm_aux += wght * nlm[:,:,ik]

    return np.sum(nlm_aux/np.sum(kwght), axis=1)


  def DR ( self, Dk, kwght ):

    nawf = Dk.shape[0]
    nkpnts = kwght.shape[0]

    D = np.zeros((nawf,nawf), dtype=complex)
    for ik,wght in enumerate(kwght):
      D += wght * Dk[:,:,ik]

    return D.real/np.sum(kwght)


  def hartree_energy ( self, DR_up, DR_dn, basis, basis_2e ):
    import itertools

    tmp_U,tmp_J = 0,0
    ind = list(itertools.product(basis_2e,repeat=4))

    for k,l,m,n in ind:
      int_U = self.coulomb(basis[m], basis[n], basis[k], basis[l])
      int_J = self.coulomb(basis[m], basis[k], basis[n], basis[l])

      a_b = DR_up[m,n]*DR_up[k,l] + DR_dn[m,n]*DR_dn[k,l]
      ab_ba = DR_dn[m,n]*DR_up[k,l] + DR_up[m,n]*DR_dn[k,l]

      tmp_U += int_U * (a_b + ab_ba)
      tmp_J += int_J * a_b

    return tmp_U, tmp_J

  def read_ham_data ( self, fpath, nspin ):
    from os.path import join

    kpnts = np.loadtxt(open(join(fpath,'k.txt'),'r'))
    kwght = np.loadtxt(open(join(fpath,'wk.txt'),'r'))

    if len(kpnts.shape) == 1:
      kpnts = np.array([kpnts])
      kwght = np.array([kwght])
    nkpnts = kpnts.shape[0]

    kovp = np.load(join(fpath,'kovp.npy'))
    nbasis = int(np.sqrt(kovp.shape[0]/nkpnts))
    kovp = kovp.reshape((nbasis,nbasis,nkpnts))

    kham_up = kham_dn = None
    for ispin in range(nspin):

      fname = 'kham'
      if nspin == 2:
        fname += '_up' if ispin==0 else '_dn'
      fname += '.npy'

      kham = np.load(join(fpath,fname))
      kham = kham.reshape((nbasis,nbasis,nkpnts))

      if ispin == 0:
        kham_up = kham
      elif ispin == 1:
        kham_dn = kham

    return kpnts, kwght, kovp, kham_up, kham_dn


  def coulomb ( self, a, b, c, d ):
    from .defs.pyints import contr_coulomb
    ' Coulomb interaction between four contracted Gaussians '
    return contr_coulomb(a.pexps,a.pcoefs,a.pnorms,a.origin,a.powers,
                         b.pexps,b.pcoefs,b.pnorms,b.origin,b.powers,
                         c.pexps,c.pcoefs,c.pnorms,c.origin,c.powers,
                         d.pexps,d.pcoefs,d.pnorms,d.origin,d.powers)



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def plot_compare_PAO_DFT_eigs(Hks,Sks,my_eigsmat,read_S):
  import numpy as np
  from scipy import linalg as spl
  from numpy import linalg as npl
  from matplotlib import pyplot as plt

  nawf,nawf,nkpnts,nspin = Hks.shape
  E_k = np.zeros((nawf,nkpnts,nspin))

  ispin = 0 #plots only 1 spin channel
  #for ispin in xrange(nspin):
  for ik in range(nkpnts):
      if read_S:
          eigval,_ = spl.eigh(Hks[:,:,ik,ispin],Sks[:,:,ik],lower=False)
      else:
          eigval,_ = npl.eigh(Hks[:,:,ik,ispin],UPLO='U')
      E_k[:,ik,ispin] = np.sort(np.real(eigval))

  fig=plt.figure
  nbnds_dft,_,_=my_eigsmat.shape
  for i in range(nbnds_dft):
      #print("{0:d}".format(i))
      yy = my_eigsmat[i,:,ispin]
      if i==0:
          plt.plot(yy,'ok',markersize=3,markeredgecolor='lime',markerfacecolor='lime',label='DFT')
      else:
          plt.plot(yy,'ok',markersize=3,markeredgecolor='lime',markerfacecolor='lime')

  for i in range(nawf):
      yy = E_k[i,:,ispin]
      if i==0:
          plt.plot(yy,'ok',markersize=2,markeredgecolor='None',label='PAO')
      else:
          plt.plot(yy,'ok',markersize=2,markeredgecolor='None')

  plt.xlabel('k-points')
  plt.ylabel('Energy - E$_F$ (eV)')
  plt.legend()
  plt.title('Comparison of PAO vs. DFT eigenvalues')
  plt.savefig('comparison.pdf',format='pdf')
  return()



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

from numpy import ndarray

def struct_from_outputfile_QE ( fname:str ):
  '''
  '''
  from os.path import isfile,join
  import numpy as np

  if not isfile(fname):
    msg = 'File {} does not exist.'.format(join(getcwd(),fname))
    raise FileNotFoundError(msg)

  struct = {'lunit':'bohr', 'aunit':'alat'}
  with open(fname, 'r') as f:
    lines = f.readlines()

    eL = 0
    nL = len(lines)

    try:
      struct['species'] = []
      celldm = np.empty(6, dtype=float)
      while 'bravais-lattice' not in lines[eL]:
        eL += 1
      ibrav = int(lines[eL].split()[3])

      while  'celldm' not in lines[eL]:
        eL += 1
      celldm[:3] = [float(v) for i,v in enumerate(lines[eL].split()) if i%2==1]
      celldm[3:] = [float(v) for i,v in enumerate(lines[eL+1].split()) if i%2==1]

      if ibrav != 0:
        from .lattice_format import lattice_format_QE
        struct['lattice'] = lattice_format_QE(ibrav, celldm)
      else:
        while 'crystal axes' not in lines[eL]:
          eL += 1
        coord = []
        for l in lines[eL+1:eL+4]:
          coord.append([celldm[0]*float(v) for v in l.split()[3:6]])
        struct['lattice'] = np.array(coord)

      while 'site n.' not in lines[eL]:
        eL += 1
      eL += 1
      apos = []
      while 'End' not in lines[eL] and lines[eL] != '\n':
        line = lines[eL].split()
        struct['species'].append(line[1])
        apos.append([float(v) for v in line[6:9]])
        eL += 1
      apos = celldm[0] * np.array(apos)
      struct['abc'] = apos @ np.linalg.inv(struct['lattice'])

    except Exception as e:
      print('ERROR: Could not read the QE output.')
      raise e

  return struct


def read_relaxed_coordinates_QE ( fname:str ):
  '''
    Reads relaxed atomic positions from a QE .out file. If vcrelax is set True, the crystal coordinates are also read.

    Arguments:
      fname (str): File name (including path) for the .out file
      vcrelax (bool): True reads crystal coordinates in addition to atomic positions
      read_all (bool): True forces all relax steps to be read. If EoF is encountered before 'final coordinates' the last coordinates to appear in the file are retunred. If no coordinates are found, an empty dictionary is returned.

    Returns:
      (dict): Dictionary with one or two entries - 'apos' for atomic positions and 'coord' for crystal coordinates.
  '''
  from os.path import isfile,join
  from os import getcwd
  import numpy as np
  import re

  abc = []
  cell_params = []
  struct = struct_from_outputfile_QE(fname)

  with open(fname, 'r') as f:
    lines = f.readlines()

    eL = 0
    nL = len(lines)

    try:
      def read_apos ( sind ):
        apos = []
        while lines[sind] != '\n' and not 'End final coordinates' in lines[sind]:
          apos.append([float(v) for v in lines[sind].split()[1:4]])
          sind += 1
        return sind, apos

      while eL < nL:
        while eL < nL and 'CELL_PARAMETERS' not in lines[eL] and 'ATOMIC_POSITIONS' not in lines[eL]:
          eL += 1
        if eL >= nL:
          break
        if 'ATOMIC_POSITIONS' in lines[eL]:
          unit = lines[eL].split()[1].strip('(){{}}')
          if len(unit) > 1:
            struct['aunit'] = unit
          eL,apos = read_apos(eL+1)
          abc.append(apos)
        elif 'CELL_PARAMETERS' in lines[eL]:
          coord = []
          unit = lines[eL].split()[1].strip('(){{}}')

          alat = 1
          if 'alat' in unit or len(unit) == 0:
            struct['lunit'] = 'alat'
            if 'alat' in unit:
              cpattern = re.search('\(([^\)]+)\)', lines[eL])
              if cpattern is not None:
                alat = float(cpattern.group(0)[1:-1].split('=')[1])
          else:
            struct['lunit'] = unit
          for l in lines[eL+1:eL+4]:
            coord.append(alat*np.array([float(v) for v in l.split()]))
          cell_params.append(coord)
          eL += 4

          while 'ATOMIC_POSITIONS' not in lines[eL]:
            eL += 1
          eL,apos = read_apos(eL+1)
          abc.append(apos)

    except Exception as e:
      print('WARNING: No atomic positions or cell coordinates were found.', flush=True)
      raise e

  struct['lattice'] = np.array([struct['lattice']] + cell_params)
  struct['abc'] = np.array([struct['abc']] + abc)

  return struct


def struct_from_inputfile_QE ( fname:str ) -> dict:
  '''
    Generate a dictionary containing all atomic information from a QE inputfile
    WARNING: Currently only the control blocks are read. Atomic cards are not...

    Arguments:
      fname (str): Name (including path) of the inputfile

    Returns:
      (dict): Structure dictionary
  '''
  from os.path import isfile
  import numpy as np
  import re

  if not isfile(fname):
    raise FileNotFoundError('File {} does not exist.'.format(fname))

  fstr = None
  with open(fname, 'r') as f:
    fstr = f.read()

  # Datatype format helpers for QE input
  nocomma = lambda s : s.replace(',', '')
  qebool = lambda s : True if s.split('.')[1][0].lower() == 't' else False
  qenum = lambda s : s.split('=')[1].replace('d', 'e')
  qeint = lambda s : int(qenum(s))
  qefloat = lambda s : float(qenum(s))
  def inquote ( s ):
    v = '"' if '"' in s else "'"
    return s.split(v)[1]

  # Process blocks
  cards = {}
  blocks = {}
  natom = ntype = 0
  pattern = re.compile('&(.*?)/@')
  celldm = np.zeros(6, dtype=float)
  comment = lambda v : v != '' and v[0] != '!'
  matches = pattern.findall(fstr.replace(' ', '').replace('\n', '@ '))
  for match in matches:

    match = match.replace(',@', '@')
    match = [s.replace(' ', '').split('!')[0] for s in re.split(', |@',match) if s!='']

    # Split inline commas without destroying Hubbard_occ tags
    block_args = []
    for m in match:

      hcinds = set([s.end(0)-1 for s in list(re.finditer('\(([^\)]+),',m))])

      if len(hcinds) == 0:
        for s in m.split(','):
          if s != '':
            block_args.append(s)

      else:

        cinds = set([i for i,c in enumerate(m) if c==','])
        if cinds == hcinds:
          block_args.append(m)

        else:
          iprev = 0
          for i in sorted(cinds.difference(hcinds)):
            block_args.append(m[iprev:i])
            iprev = i+1
          block_args.append(m[iprev:])

    match = None
    block = block_args.pop(0).lower()
    mf = filter(comment, block_args)

    blocks[block] = {}
    for s in block_args:
      k,v = s.split('=')
      k = k.lower()
      blocks[block][k] = v
      if k == 'ntyp':
        ntype = int(v)
      elif k == 'nat':
        natom = int(v)

  # Process CARDS
  fstr = list(filter(comment, fstr.split('\n')))
  def scan_blank_lines ( nl ):
    nl += 1
    while fstr[nl] == '':
      nl += 1
    return nl

  il = 0
  nf = len(fstr)
  while il < nf and 'ATOMIC_POSITIONS' not in fstr[il]:
    il += 1
  if il < nf:
    cards['ATOMIC_POSITIONS'] = [fstr[il]]
    il = scan_blank_lines(il)
    for i in range(natom):
      cards['ATOMIC_POSITIONS'].append(fstr[il+i])

  sl = 0
  while sl < nf and 'ATOMIC_SPECIES' not in fstr[sl]:
    sl += 1
  if sl < nf:
    cards['ATOMIC_SPECIES'] = [fstr[sl]]
    sl = scan_blank_lines(sl)
    for i in range(ntype):
      cards['ATOMIC_SPECIES'].append(fstr[sl+i])

  hl = 0
  while hl < nf and 'HUBBARD' not in fstr[hl]:
    hl += 1
  if hl < nf:
    cards['HUBBARD'] = [fstr[hl]]
    hl = scan_blank_lines(hl)
    for i in range(5):
      if fstr[hl+i].split()[0] == 'U':
        cards['HUBBARD'].append(fstr[hl+i])
      else:
        break

  kl = 0
  while kl < nf and 'K_POINTS' not in fstr[kl]:
    kl += 1
  if kl < nf:
    cards['K_POINTS'] = [fstr[kl]]
    if 'gamma' in fstr[kl].lower():
      pass
    else:
      cards['K_POINTS'].append(fstr[kl+1])
      if 'automatic' not in fstr[kl]:
        nk = int(fstr[kl+1])
        kl += 2
        for i in range(nk):
          cards['K_POINTS'].append(fstr[kl+i])

  cl = 0
  while cl < nf and 'CELL_PARAM' not in fstr[cl]:
    cl += 1
  if cl < nf:
    cards['CELL_PARAMETERS'] = []
    for i in range(4):
      cards['CELL_PARAMETERS'].append(fstr[cl+i])

  return blocks, cards


def create_atomic_inputfile ( calculation, blocks, cards ):

  with open(f'{calculation}.in', 'w') as f:
    f.write('\n')
    for kb,vb in blocks.items():
      f.write(f' &{kb}\n')
      for ks,vs in vb.items():
        f.write(f'  {ks} = {vs}\n')
      f.write(' /\n\n')

    if 'ATOMIC_SPECIES' in cards:
      for s in cards['ATOMIC_SPECIES']:
        f.write(s + '\n')
      f.write('\n')
      del cards['ATOMIC_SPECIES']

    for kc,vc in cards.items():
      for s in vc:
        f.write(s + '\n')
      f.write('\n')


def create_acbn0_inputfile ( prefix, pthr ):

  with open('acbn0.py', 'w') as f:
    f.write('from PAOFLOW import PAOFLOW\n\n')
    f.write(f'paoflow = PAOFLOW.PAOFLOW(savedir=\'{prefix}.save\', acbn0=True)\n')
    f.write('paoflow.read_atomic_proj_QE()\n')
    f.write(f'paoflow.projectability(pthr={pthr})\n')
    f.write('paoflow.pao_hamiltonian(write_binary=True)\n')
    f.write('paoflow.finish_execution()\n')



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

# this version works only for non-magnetic or non-collienar calculations
def wave_function_site_projection(data_controller):
    import cmath
    import numpy as np
    from mpi4py import MPI
    from os.path import join
    from scipy import fftpack as FFT
    from .constants import ANGSTROM_AU
    from scipy.fftpack import fftshift
    from .communication import scatter_full,gather_full
    
    arry,attr = data_controller.data_dicts()

    tau = arry['tau'] / ANGSTROM_AU
    naw,v_k = arry['naw'],arry['v_k']   
    bands,k_index = arry['bands_proj'],attr['k_proj']   
    
    # sites to project the wave function
    site = np.copy(tau[:,0])

    do_spin_orbit = attr['do_spin_orbit']    
    nawf,dim = attr['nawf'],attr['dimension']

    for idb in range(len(bands)):
        bnd_idx = bands[idb]  # index of the band to be projected
        # open file
        f = open(join(attr['opath'],'site-projected-wave-function-'+str(bnd_idx)+'.dat'), 'w')
        for n in range(tau.shape[0]):

            # Do to the doubling of the Hamiltonian when SOC is included in the PAO Hamiltonian.
            if (do_spin_orbit):
                # creating masks to consirer only the n site.
                # seting up the nonzero parts of the mask and the wave-function
                idx= int(np.sum(naw[0:n])) # initial
                fdx= int(idx + naw[n])     # final
                s = int(nawf/2)

                usector_idx = np.arange(idx,fdx,dtype=int)
                dsector_idx = np.arange(idx+s,fdx+s,dtype=int)
                idx_list = list(np.append(usector_idx,dsector_idx))

                total=0
                total+=np.sum(np.absolute(np.square(v_k[k_index:k_index+1,idx_list,bnd_idx,0])))

            else: # no SOC or SOC form QE.
                # creating masks to consirer only the n site.
                # seting up the nonzero parts of the mask and the wave-function
                idx= int(np.sum(naw[0:n])) # initial
                fdx= int(idx + naw[n])     # final

                total=0
                total+=np.sum(np.absolute(np.square(v_k[k_index:k_index+1,idx:fdx,bnd_idx,0])))

            if(dim==3): # ploting for 3D system
                # we sum a very small part 0.0001 for ploting purpose.
                f.write (("%5.4f %5.4f %5.4f %5.4f \n") %(tau[n,0],tau[n,1],tau[n,2],total+0.0001))
            if(dim==2): # ploting for 2D system
                # we sum a very small part 0.0001 for ploting purpose.
                f.write (("%5.4f %5.4f %5.4f  \n") %(tau[n,0],tau[n,1],total+0.0001))
            if(dim==1): # ploting for 1D system
                # we sum a very small part 0.0001 for ploting purpose.
                f.write (("%5.4f %5.4f  \n") %(tau[n,2],total+0.0001))

        f.close()




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import xml.etree.cElementTree as ET
from mpi4py import MPI
import numpy as np

def parse_qe_data_file_schema ( data_controller, fname ):
  '''
  Parse the data_file_schema.xml file produced by Quantum Espresso.
  Populated the DataController object with all necessay information.

  Arugments:
    data_controller (DataController): Data controller to populate
    fname (str): Path and name of the xml file.
  '''
  import re

  arry,attr = data_controller.data_dicts()
  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  verbose = attr['verbose']

  Hart2eV = 2 * 13.60569193

  tree = ET.parse(fname)
  root = tree.getroot()

  qe_version = float(re.findall('\d+\.\d+', root.find('general_info/creator').attrib['VERSION'])[0])

  dftSO = True if root.find('input/spin/spinorbit').text=='true' else False

  elem = root.find('output')

  species,pseudos = [],[]
  lspecies = elem.findall('atomic_species/species')
  for s in lspecies:
    species.append(s.attrib['name'])
    pseudos.append(s.find('pseudo_file').text)

  astruct = elem.find('atomic_structure')
  alat = float(astruct.attrib['alat'])
  natoms = int(astruct.attrib['nat'])

  basis_elem = elem.find('basis_set')
  ecutwfc = float(basis_elem.find('ecutwfc').text)*2
  ecutrho = float(basis_elem.find('ecutrho').text)*2 # it is read in Ha and must be converted to Ry

  a_vectors,b_vectors = [],[]
  vec_array = lambda n,e,s : np.array(e.find('%s%d'%(s,n)).text.split(), dtype=float)
  for i in range(1, 4):
    a_vectors.append(vec_array(i,astruct,'cell/a'))
    b_vectors.append(vec_array(i,basis_elem,'reciprocal_lattice/b'))
  a_vectors = np.array(a_vectors)/alat
  b_vectors = np.array(b_vectors)

  atoms = []
  tau = np.empty((natoms,3), dtype=float)
  latoms = elem.findall('atomic_structure/atomic_positions/atom')
  for i in range(natoms):
    atoms.append(latoms[i].attrib['name'].strip())
    tau[i,:] = np.array(latoms[i].text.split(), dtype=float)

  insulator = True
  smearing = elem.find('band_structure/smearing')
  if smearing is not None:
    insulator = False

  dftMag = False
  mag_elem = elem.find('magnetization/do_magnetization')
  if mag_elem is not None:
    if mag_elem.text == 'true':
      dftMag = True

  mpg = elem.find('band_structure/starting_k_points/monkhorst_pack').attrib
  k1,k2,k3 = int(mpg['k1']),int(mpg['k2']),int(mpg['k3'])
  nk1,nk2,nk3 = int(mpg['nk1']),int(mpg['nk2']),int(mpg['nk3'])
  if verbose: print('Monkhorst and Pack grid:',nk1,nk2,nk3,k1,k2,k3)

  bs = elem.find('band_structure')
  nkpnts = int(bs.find('nks').text)
  nelec = int(float(bs.find('nelec').text))
  nawf = int(bs.find('num_of_atomic_wfc').text)
  lsda = True if bs.find('lsda').text=='true' else False
  nspin = 2 if lsda else 1

  if nspin == 1:
    nbnds = int(bs.find('nbnd').text)
  else:
    nb_up = int(bs.find('nbnd_up').text)
    nb_dw = int(bs.find('nbnd_dw').text)
    if nb_up != nb_dw and verbose: print('nbnd_up (%d) != nbnd_dw (%d)'%(nbnd_up,nbnd_dw))
    nbnds = np.max([nb_up, nb_dw])

  if rank == 0 and verbose:
    print('Number of kpoints: {0:d}'.format(nkpnts))
    print('Number of electrons: {0:d}'.format(nelec))
    print('Number of bands: {0:d}'.format(nbnds))

  ef_text = None
  try:
    ef_text = bs.find('fermi_energy').text
  except:
    try:
      ef_text = bs.find('highestOccupiedLevel').text
    except Exception as e:
      print('Fermi energy not located in QE data file.')
      raise e
  Efermi = float(ef_text) * Hart2eV

  kpnts = np.empty((nkpnts,3), dtype=float)
  kpnt_weights = np.empty((nkpnts), dtype=float)
  eigs = np.empty((nbnds,nkpnts,nspin), dtype=float)

  kse = bs.findall('ks_energies')
  for i,k in enumerate(kse):
    kp = k.find('k_point')
    kpnts[i,:] = np.array(kp.text.split(), dtype=float)
    kpnt_weights[i] = float(kp.attrib['weight'])
    eigs[:,i,:] = np.array(k.find('eigenvalues').text.split(), dtype=float).reshape((nspin,nbnds)).T
  eigs = eigs * Hart2eV - Efermi

  sym_rot,shifts = [],[]
  eq_atoms,sym_info,time_rev = [],[],[]

  lsyms = elem.findall('symmetries/symmetry')
  for i,sym in enumerate(lsyms):
    info = sym.find('info')
    if info.text == 'crystal_symmetry':
      sym_info.append(info.attrib['name'])

      shift_text = sym.find('fractional_translation').text
      shifts.append([float(v) for v in shift_text.split()])

      eq_atom_text = sym.find('equivalent_atoms').text
      eq_atoms.append([int(v) for v in eq_atom_text.split()])

      if 'time_reversal' in info.attrib:
        time_rev.append(info.attrib['time_reversal'])

      rots = []
      for rot in sym.find('rotation').text.split('\n'):
        line = rot.split()
        if len(line) != 0:
          rots.append([float(v) for v in rot.split()])
      sym_rot.append(rots)

  shifts = np.array(shifts)
  eq_atoms = np.array(eq_atoms) - 1
  sym_info = np.array(sym_info)

  sym_rot = np.transpose(np.array(sym_rot), axes=(0,2,1))

  if len(time_rev) == 0:
    time_rev = np.zeros(sym_info.shape[0], dtype=bool)
  else:
    time_rev = np.array([True if v=='true' else False for v in time_rev])

  omega = alat**3 * a_vectors[0,:].dot(np.cross(a_vectors[1,:],a_vectors[2,:]))

  attrs = [('qe_version',qe_version),('alat',alat),('nk1',nk1),('nk2',nk2),('nk3',nk3),('ok1',k1),('ok2',k2),('ok3',k3),('natoms',natoms),('ecutrho',ecutrho),('ecutwfc',ecutwfc),('nawf',nawf),('nbnds',nbnds),('nspin',nspin),('nkpnts',nkpnts),('nelec',nelec),('Efermi',Efermi),('omega',omega),('dftSO',dftSO),('dftMAG',dftMag),('insulator',insulator)]
  for s,v in attrs:
    attr[s] = v

  spec = [(species[i],pseudos[i]) for i in range(len(species))]
  arrys = [('tau',tau),('atoms',atoms),('species',spec),('a_vectors',a_vectors),('b_vectors',b_vectors),('equiv_atom',eq_atoms),('kpnts',kpnts),('kpnts_wght',kpnt_weights),('my_eigsmat',eigs),('sym_rot',sym_rot),('sym_shift',shifts),('sym_info',sym_info),('sym_TR',time_rev)]
  for s,v in arrys:
    arry[s] = v


def parse_qe_data_file ( data_controller, fpath, fname ):
  '''
  Parse the data_file.xml file produced by earlier versions of Quantum Espresso.
  Populated the DataController object with all necessay information.

  Arugments:
    data_controller (DataController): Data controller to populate
    fname (str): Path and name of the xml file.
  '''
  import re
  from os.path import join

  arry,attr = data_controller.data_dicts()
  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  verbose = attr['verbose']

  Hart2eV = 2 * 13.60569193

  tree = ET.parse(join(fpath,fname))
  root = tree.getroot()

  qe_version = float(re.findall('\d+\.\d+', root.find('HEADER/CREATOR').attrib['VERSION'])[0])

  dftSO = root.find('SPIN/SPIN-ORBIT_CALCULATION').text.replace('\n','').strip()
  dftSO = True if dftSO=='T' else False

  species,pseudos = [],[]
  alat = float(root.find('CELL/LATTICE_PARAMETER').text)
  nspec = int(root.find('IONS/NUMBER_OF_SPECIES').text)
  natoms = int(root.find('IONS/NUMBER_OF_ATOMS').text)

  for i in range(1,nspec+1):
    spec_elem = root.find('IONS/SPECIE.%d'%i)
    species.append(spec_elem.find('ATOM_TYPE').text.replace('\n','').strip())
    pseudos.append(spec_elem.find('PSEUDO').text.replace('\n','').split()[0])

  mpg = root.find('BRILLOUIN_ZONE/MONKHORST_PACK_OFFSET').attrib
  k1,k2,k3 = int(mpg['k1']),int(mpg['k2']),int(mpg['k3'])

  mpg = root.find('BRILLOUIN_ZONE/MONKHORST_PACK_GRID').attrib
  nk1,nk2,nk3 = int(mpg['nk1']),int(mpg['nk2']),int(mpg['nk3'])

  if verbose: print('Monkhorst and Pack grid:',nk1,nk2,nk3,k1,k2,k3)

  ecutrho = float(root.find('PLANE_WAVES/RHO_CUTOFF').text.strip())

  a_vectors,b_vectors = [],[]
  a_elem = root.find('CELL/DIRECT_LATTICE_VECTORS')
  b_elem = root.find('CELL/RECIPROCAL_LATTICE_VECTORS')
  vec_array = lambda n,e,s : np.array(e.find('%s%d'%(s,n)).text.split(), dtype=float)
  for i in range(1, 4):
    a_vectors.append(vec_array(i,a_elem,'a'))
    b_vectors.append(vec_array(i,b_elem,'b'))
  a_vectors = np.array(a_vectors)/alat
  b_vectors = np.array(b_vectors)

  atoms = []
  tau = np.empty((natoms,3), dtype=float)
  for i in range(natoms):
    latom = root.find('IONS/ATOM.%d'%(i+1))
    atoms.append(latom.attrib['SPECIES'].strip())
    tau[i,:] = np.array(latom.attrib['tau'].split(), dtype=float)

  insulator = False if root.find('OCCUPATIONS/SMEARING_METHOD').text.strip()=='T' else True

  if root.find('MAGNETIZATION_INIT').text.strip() == 'T':
    raise NotImplementedError('Two Fermi energies unhandled')

  bs = root.find('BAND_STRUCTURE_INFO')
  nkpnts = int(bs.find('NUMBER_OF_K-POINTS').text.strip())
  nawf = int(bs.find('NUMBER_OF_ATOMIC_WFC').text.strip())
  nbnds = int(bs.find('NUMBER_OF_BANDS').text.strip())
  nelec = int(float(bs.find('NUMBER_OF_ELECTRONS').text.strip()))
  Efermi = float(bs.find('FERMI_ENERGY').text.strip()) * Hart2eV
  lsda = True if root.find('SPIN/LSDA').text.strip()=='T' else False
  nspin = 2 if lsda else 1

  if rank == 0 and verbose:
    print('Number of kpoints: {0:d}'.format(nkpnts))
    print('Number of electrons: {0:d}'.format(nelec))
    print('Number of bands: {0:d}'.format(nbnds))

  kpnts = np.empty((nkpnts,3), dtype=float)
  kpnt_weights = np.empty((nkpnts), dtype=float)
  eigs = np.empty((nbnds,nkpnts,nspin), dtype=float)

  ev = root.find('EIGENVALUES')
  data_tags = ['DATAFILE'] if nspin==1 else ['DATAFILE.1', 'DATAFILE.2']
  for i in range(nkpnts):
    k = ev.find('K-POINT.%d'%(i+1))
    kpnts[i,:] = np.array(k.find('K-POINT_COORDS').text.split(), dtype=float)
    kpnt_weights[i] = float(k.find('WEIGHT').text)
    for ispin,ftag in enumerate(data_tags):
      efname = k.find(ftag).attrib['iotk_link']
      eroot = ET.parse(join(fpath,efname)).getroot()
      eigs[:,i,ispin] = np.array(eroot.find('EIGENVALUES').text.split(), dtype=float)
  eigs = eigs * Hart2eV - Efermi

  dftMag = False
  if root.find('SPIN/SPIN-ORBIT_DOMAG').text.replace('\n','').strip() == 'T':
    dftMag = True

  sym_rot,shifts = [],[]
  eq_atoms,sym_info,time_rev = [],[],[]

  lsyms = root.find('SYMMETRIES')
  nsyms = int(lsyms.find('NUMBER_OF_SYMMETRIES').text)
  for i in range(1,nsyms+1):
    sym = lsyms.find('SYMM.%d'%i)

    shift_text = sym.find('FRACTIONAL_TRANSLATION').text
    shifts.append([float(v) for v in shift_text.split()])

    eq_atom_text = sym.find('EQUIVALENT_IONS').text
    eq_atoms.append([int(v) for v in eq_atom_text.split()])

    info = sym.find('INFO')
    sym_info.append(info.attrib['NAME'])
    time_rev.append(int(info.attrib['T_REV']))

    rots = []
    for rot in sym.find('ROTATION').text.split('\n'):
      line = rot.split()
      if len(line) != 0:
        rots.append([float(v) for v in line])
    sym_rot.append(rots)

  shifts = np.array(shifts)
  eq_atoms = np.array(eq_atoms) - 1

  sym_rot = np.transpose(np.array(sym_rot), axes=(0,2,1))

  sym_info = np.array(sym_info)
  time_rev = np.array(time_rev, dtype=bool)

  # Compute the cell volume
  omega = alat**3 * a_vectors[0,:].dot(np.cross(a_vectors[1,:],a_vectors[2,:]))

  # Add the attributes and arrays to the data controller
  attrs = [('qe_version',qe_version),('alat',alat),('nk1',nk1),('nk2',nk2),('nk3',nk3),('ok1',k1),('ok2',k2),('ok3',k3),('natoms',natoms),('ecutrho',ecutrho),('nawf',nawf),('nbnds',nbnds),('nspin',nspin),('nkpnts',nkpnts),('nelec',nelec),('Efermi',Efermi),('omega',omega),('dftSO',dftSO),('dftMAG',dftMag),('insulator',insulator)]
  for s,v in attrs:
    attr[s] = v

  spec = [(species[i],pseudos[i]) for i in range(len(species))]
  arrys = [('tau',tau),('atoms',atoms),('species',spec),('a_vectors',a_vectors),('b_vectors',b_vectors),('equiv_atom',eq_atoms),('kpnts',kpnts),('kpnts_wght',kpnt_weights),('my_eigsmat',eigs),('sym_rot',sym_rot),('sym_shift',shifts),('sym_info',sym_info),('sym_TR',time_rev)]
  for s,v in arrys:
    arry[s] = v


def parse_qe_atomic_proj ( data_controller, fname ):
  '''
  Parse the atomic_proj.xml file produced by Quantum Espresso.
  Populated the DataController object with all necessay information.

  Arugments:
    data_controller (DataController): Data controller to populate
    fname (str): Path and name of the xml file.
  '''

  arry,attr = data_controller.data_dicts()
  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  verbose = attr['verbose']
  acbn0 = attr['acbn0']

  tree = ET.parse(fname)
  root = tree.getroot()
  qe_version = attr['qe_version']

  Ry2eV = 13.60569193
  Efermi = attr['Efermi']

  elem = root.find('HEADER')
  if qe_version > 6.5:
    header = elem.attrib
    nkpnts = int(header['NUMBER_OF_K-POINTS'])
    nspin = int(header['NUMBER_OF_SPIN_COMPONENTS'])
    nbnds = int(header['NUMBER_OF_BANDS'])
    nawf = int(header['NUMBER_OF_ATOMIC_WFC'])
  else:
    nbnds = int(elem.find('NUMBER_OF_BANDS').text)
    nkpnts = int(elem.find('NUMBER_OF_K-POINTS').text)
    nspin = int(elem.find('NUMBER_OF_SPIN_COMPONENTS').text)
    nawf = int(elem.find('NUMBER_OF_ATOMIC_WFC').text)

  if nspin == 4:
    nspin = 1

  wavefunctions = np.empty((nbnds,nawf,nkpnts,nspin), dtype=complex)
  overlaps = np.empty((nawf,nbnds,nkpnts), dtype=complex) if acbn0 else None

  if qe_version > 6.5:

    elem = root.find('EIGENSTATES')
    ewfc = elem.findall('PROJS')
    for ispin in range(nspin):
      for i in range(nkpnts):
        ik = ispin*nkpnts + i
        for j,proj in enumerate(ewfc[ik].findall('ATOMIC_WFC')):
          ind = int(proj.attrib['index'])-1
          spin = int(proj.attrib['spin'])-1
          text = [float(v) for v in proj.text.split()]
          for k in range(len(text)//2):
            k2 = 2*k
            wavefunctions[k,ind,i,ispin] = complex(text[k2],text[k2+1])

    if acbn0:
      elem = root.find('OVERLAPS')
      for i,ovp in enumerate(elem.findall('OVPS')):
        dim = int(ovp.attrib['dim'])
        ispin = int(ovp.attrib['spin'])-1
        text = ovp.text.split()
        for j in range(len(text)//2):
          j2 = 2*j
          i1 = j//dim
          i2 = j%dim
          v1,v2 = float(text[j2]),float(text[j2+1])
          overlaps[i1,i2,i] = complex(v1,v2)

  else:

    def read_wf ( kpnt, i, ispin ):
      for j,wf in enumerate(kpnt):
        text = wf.text.replace(',', ' ').split()
        wft = np.array(text, dtype=float).reshape((nbnds,2))
        wavefunctions[:,j,i,ispin] = wft[:,0] + 1j*wft[:,1]

    elem = root.find('PROJECTIONS')
    for i,kpnt in enumerate(elem):
      if nspin == 1:
        read_wf(kpnt, i , 0)
      else:
        for ispin in range(nspin):
          nele = kpnt.find('SPIN.%d'%(ispin+1))
          read_wf(nele, i, ispin)

    if acbn0:
      elem = root.find('OVERLAPS')
      for i,kpnt in enumerate(elem):
        for j,ovp in enumerate(kpnt):
          ispin = int(ovp.tag.split('.')[1]) - 1
          text = ovp.text.replace(',', ' ').split()
          for k in range(len(text)//2):
            k0,k1 = k//nbnds,k%nbnds
            k2 = 2*k
            v1,v2 = float(text[k2]),float(text[k2+1])
            overlaps[k0,k1,i] = complex(v1,v2)

  arrys = [('U',wavefunctions)]
  if acbn0:
    arrys += [('Sks',overlaps)]
  for s,v in arrys:
    arry[s] = v



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
import xml.etree.ElementTree as ET
from io import StringIO

# read UPF utility from Davide Ceresoli
class UPF:

  def __init__ ( self, filename ):
    """Open a UPF file, determine version, and read it"""
    with open(filename) as f:
      xml_file_content = f.read()

    # fix broken XML
    xml_file_content = xml_file_content.replace('&', '&amp;')

    # test if v1
    if xml_file_content.startswith('<PP_INFO>'):
      xml_file_content = '<UPF version="1.0">\n' + xml_file_content + '</UPF>\n'

    # parse the XML file
    root = ET.fromstring(xml_file_content)

    # dispatch to the specific routine
    upfver = root.attrib["version"]
    self.version = int(upfver.split(".")[0])
    if self.version == 1:
      self._read_upf_v1(root)
    elif self.version == 2:
      self._read_upf_v2(root)
    else:
      raise RuntimeError('Wrong UPF version: %s'%upfver)


  def _read_upf_v1 ( self, root ):
    """Read a UPF v1 pseudopotential"""

    # parse info and header
    self.info = root.find('PP_INFO').text
    for line in root.find('PP_HEADER').text.split('\n'):
      l = line.split()
      if 'Element' in line:       self.element = l[0]
      if 'NC' in line:            self.ptype = 'NC'
      if 'US' in line:            self.ptype = 'US'
      if 'Nonlinear' in line:     self.nlcc = l[0] == 'T'
      if 'Exchange' in line:      self.qexc = ' '.join(l[0:4])
      if 'Z valence' in line:     self.val = float(l[0])
      if 'Max angular' in line:   self.lmax = int(l[0])
      if 'Number of po' in line:  self.npoints = int(l[0])
      if 'Number of Wave' in line:
        self.nwfc = int(l[0])
        self.nproj = int(l[1])

    # parse mesh
    text = root.find('PP_MESH/PP_R').text
    self.r = np.array( [float(x) for x in text.split()] )
    text = root.find('PP_MESH/PP_RAB').text
    self.rab = np.array( [float(x) for x in text.split()] )

    # local potential
    text = root.find('PP_LOCAL').text
    self.vloc = np.array( [float(x) for x in text.split()] ) / 2.0  # to Hartree

    # atomic wavefunctions
    self.pswfc = []
    chis = root.find('PP_PSWFC')
    self.shells = []
    self.jchia = []
    if chis is not None:
      data = StringIO(chis.text)
      nlines = self.npoints//4
      if self.npoints % 4 != 0: nlines += 1

      while True:
        line = data.readline()
        if line == '\n': continue
        if line == '': break
        label, l, occ, dummy = line.split()
 
        wfc = []
        for i in range(nlines):
          wfc.extend(map(float, data.readline().split()))
        wfc = np.array(wfc)
        self.shells.append(int(l))
        self.pswfc.append( {'label': label, 'occ': float(occ), 'wfc': wfc} )

    # atomic rho
    self.atrho = None
    atrho = root.find('PP_RHOATOM')
    if atrho is not None:
      self.atrho = np.array( [float(x) for x in atrho.text.split()] )

    # TODO: NLCC

    # TODO: PS_NONLOCAL/BETA, PP_DIJ

    # TODO: GIPAW data


  def _read_upf_v2 ( self, root ):
    """Read a UPF v2 pseudopotential"""

    # parse header
    h = root.find('PP_HEADER').attrib
    self.element = h['element']
    self.type = h['pseudo_type']
    self.nlcc = h['core_correction'] == 'true'
    self.qexc = h['functional']
    self.val = float(h['z_valence'])
    self.lmax = int(h['l_max'])
    self.npoints = int(h['mesh_size'])
    self.nwfc = int(h['number_of_wfc'])
    self.nproj = int(h['number_of_proj'])
    self.v2_header = h.copy()

    # parse mesh
    text = root.find('PP_MESH/PP_R').text
    self.r = np.array( [float(x) for x in text.split()] )
    text = root.find('PP_MESH/PP_RAB').text
    self.rab = np.array( [float(x) for x in text.split()] )

    # local potential
    text = root.find('PP_LOCAL').text
    self.vloc = np.array( [float(x) for x in text.split()] ) / 2.0  # to Hartree

    # atomic wavefunctions
    self.pswfc = []
    self.jchia = []
    self.shells = []
    i = 0
    while True:
      i += 1
      chi = root.find('PP_PSWFC/PP_CHI.%i' % (i))
      if chi is None: break

      label = chi.attrib["label"]
      occ = float(chi.attrib["occupation"])
      wfc = [float(x) for x in chi.text.split()]
      wfc = np.array(wfc)
      self.shells.append(int(chi.attrib['l']))
      self.pswfc.append( {'label': label, 'occ': float(occ), 'wfc': wfc} )

      jchi = root.find('PP_SPIN_ORB/PP_RELWFC.%d'%i)
      if jchi is not None:
        self.jchia.append(float(jchi.attrib['jchi']))
    self.jchia = self.jchia
    self.shells = self.shells

    # TODO: NLCC, ATRHO

    # TODO: PS_NONLOCAL/BETA, PP_DIJ

    # TODO: GIPAW data



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def do_gradient ( data_controller ):
  import numpy as np
  from scipy import fftpack as FFT
  from .get_R_grid_fft import get_R_grid_fft

  arry,attr = data_controller.data_dicts()

  #----------------------
  # Compute the gradient of the k-space Hamiltonian
  #----------------------

  nktot = attr['nkpnts']
  snawf,nk1,nk2,nk3,nspin = arry['Hksp'].shape

  # fft grid in R shifted to have (0,0,0) in the center
  get_R_grid_fft(data_controller, nk1, nk2, nk3)

  arry['dHksp'] = np.empty((snawf,nk1,nk2,nk3,3,nspin), dtype=complex, order='C')
  for ispin in range(nspin):
    for n in range(snawf):
      ########################################
      ### real space grid replaces k space ###
      ########################################
      if attr['use_cuda']:
        arry['Hksp'][n,:,:,:,ispin] = cuda_ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']
      else:
        arry['Hksp'][n,:,:,:,ispin] = FFT.ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']

      # Compute R*H(R)
      for l in range(3):
        arry['dHksp'][n,:,:,:,l,ispin] = FFT.fftn(arry['Rfft'][:,:,:,l]*arry['Hksp'][n,:,:,:,ispin])




# 
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from scipy import signal
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_Boltz_tensors ( data_controller, smearing, temp, ene, velkp, ispin, channels, weights):
  # Compute the L_alpha tensors for Boltzmann transport

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  arrays['scattering_tau'] = get_tau(data_controller, temp, channels, weights)

#### Forced t_tensor to have all components
  t_tensor = np.array([[0,0],[1,1],[2,2],[0,1],[0,2],[1,2]], dtype=int)

  # Quick call function for L_loop (None is smearing type)
  fLloop = lambda spol : L_loop(data_controller, temp, smearing, ene, velkp, t_tensor, spol, ispin)

  # Quick call function for Zeros on rank Zero
  zoz = lambda r: (np.zeros((3,3,esize), dtype=float) if r==0 else None)

  L0 = zoz(rank)
  L0aux = fLloop(0)
  comm.Reduce(L0aux, L0, op=MPI.SUM)
  L0aux = None

  if rank == 0:
    # Assign lower triangular to upper triangular
    sym = lambda L : (L[0,1], L[0,2], L[1,2])
    L0[1,0],L0[2,0],L0[2,1] = sym(L0)

  L1 = zoz(rank)
  L1aux = fLloop(1)
  comm.Reduce(L1aux, L1, op=MPI.SUM)
  L1aux = None

  L2 = zoz(rank)
  L2aux = fLloop(2)
  comm.Reduce(L2aux, L2, op=MPI.SUM)
  L2aux = None

  if rank == 0:
    L1[1,0],L1[2,0],L1[2,1] = sym(L1)
    L2[1,0],L2[2,0],L2[2,1] = sym(L2)

  return (L0, L1, L2) if rank==0 else (None, None, None)


def do_Boltz_tensors_hall ( data_controller, smearing, temp, ene, velkp, ispin, channels, weights):

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  arrays['scattering_tau'] = get_tau(data_controller, temp, channels, weights)

#### Forced t_tensor to have all components
  t_tensor = np.array([[0,0],[1,1],[2,2],[0,1],[0,2],[1,2]], dtype=int)

  # Quick call function for L_loop (None is smearing type)

  # Quick call function for Zeros on rank Zero
  zoz = lambda r: (np.zeros((3,3,3,esize), dtype=float) if r==0 else None)

  L0_hall = zoz(rank)
  L0_hall_aux = L_loop_hall(data_controller, temp, smearing, ene, velkp, t_tensor, 0, ispin)
  comm.Reduce(L0_hall_aux, L0_hall, op=MPI.SUM)
  L0_hall_aux = None
  
  return L0_hall if rank==0 else None


def get_tau ( data_controller, temp, channels, weights ):
  import numpy as np
  import scipy.constants as cp
  from .TauModel import TauModel
  from .do_tau_models import builtin_tau_model

  arry,attr = data_controller.data_dicts()
  snktot = arry['E_k'].shape[0]

  taus = []
  for c in channels:
    if c == 'acoustic':
      a_tau = np.ones((snktot), dtype=float)
      taus.append(a_tau)

  bnd = attr['bnd']
  eigs = np.abs(arry['E_k'][:,:bnd,:])
  snktot,_,nspin = eigs.shape

  models = []
  if channels != None:
    if len(weights) == 0:
      weights = np.ones(len(channels))
    elif len(weights) != len(models):
      raise Exception('Length of weights does not match the number of channels.')
    for i,c in enumerate(channels):
      if isinstance(c,str):
        models.append(builtin_tau_model(c,attr['tau_dict'],weights[i]))
      elif isinstance(c,TauModel):
        c.weight = weights[i]
        models.append(c)
      else:
        print('Invalid channel type.')

  if len(models) == 0:
    # Constant relaxation time approximation with tau = 1
    tau = np.ones((snktot,bnd,nspin), dtype=float)

  else:
    # Compute tau as a harmonic sum of scattering channel contributions.
    tau = np.zeros((snktot,bnd,nspin), dtype=float)
    for m in models:
      try:
        tau += m.weight/m.evaluate(temp, eigs)
      except KeyError as e:
        from .report_exception import report_exception
        print('Ensure that all required parameters are specified in the provided dictionary.')
        report_exception()
        raise e
    tau = 1/tau

  return tau



def L_loop ( data_controller, temp, smearing, ene, velkp, t_tensor, alpha, ispin ):
  from .smearing import gaussian,metpax
  # We assume tau=1 in the constant relaxation time approximation

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size

  snktot = arrays['E_k'].shape[0]

  bnd = attributes['bnd']
  kq_wght = 1./attributes['nkpnts']
  if smearing is not None and smearing != 'gauss' and smearing != 'm-p':
    print('%s Smearing Not Implemented.'%smearing)
    comm.Abort()

  L = np.zeros((3,3,esize), dtype=float)

  for n in range(bnd):
    Eaux = np.reshape(np.repeat(arrays['E_k'][:,n,ispin],esize), (snktot,esize))
    delk = (np.reshape(np.repeat(arrays['deltakp'][:,n,ispin],esize), (snktot,esize)) if smearing!=None else None)
    EtoAlpha = np.power(Eaux[:,:]-ene, alpha)
    if smearing is None:
      Eaux -= ene
      smearA = 1/(4*temp*(np.cosh(Eaux/(2*temp))**2))
    else:
      if smearing == 'gauss':
        smearA = gaussian(Eaux, ene, delk)
      elif smearing == 'm-p':
        smearA = metpax(Eaux, ene, delk)
    for l in range(t_tensor.shape[0]):
      i = t_tensor[l][0]
      j = t_tensor[l][1]
      L[i,j,:] += np.sum(kq_wght*arrays['scattering_tau'][:,n,ispin]*velkp[:,i,n,ispin]*velkp[:,j,n,ispin]*(smearA*EtoAlpha).T, axis=1)
  '''
  # noise reduction using a running average (correlation function)
  # Only possible for sigma vs chemical potential
  win = int(esize*0.025)
  N = esize
  for l in range(t_tensor.shape[0]):
    i = t_tensor[l][0]
    j = t_tensor[l][1]
    L[i,j,:] = signal.correlate(L[i,j,:] , np.ones(win), mode='same', method='fft')/win
  '''
  return L

def L_loop_hall ( data_controller, temp, smearing, ene, velkp, t_tensor, alpha, ispin ):
  from scipy.constants import hbar
  from sympy import Eijk
  from .smearing import gaussian,metpax
  from os.path import join

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size

  snktot = arrays['E_k'].shape[0]
  bohr2m = 5.29177e-11
  ev2J = 1.60218e-19
  bnd = attributes['bnd']
  nspin = attributes['nspin']
  kq_wght = 1./attributes['nkpnts']
  if smearing is not None and smearing != 'gauss' and smearing != 'm-p':
    print('%s Smearing Not Implemented.'%smearing)
    comm.Abort()

  L_hall = np.zeros((3,3,3,esize), dtype=float)
  sig_hall = np.zeros((3,3,3,snktot,bnd,nspin))

  M_inv = np.zeros((3,3,snktot,bnd,nspin))
  eff_mass_inv = arrays['d2Ed2k']
  for l in range(t_tensor.shape[0]):
    i = t_tensor[l][0]
    j = t_tensor[l][1]
    if i==j:
      M_inv[i,j]=eff_mass_inv[i]
    elif i==0 and j==1:
      M_inv[i,j]=eff_mass_inv[3]
      M_inv[j,i]=eff_mass_inv[3]
    elif i==0 and j==2:
      M_inv[i,j]=eff_mass_inv[4]
      M_inv[j,i]=eff_mass_inv[4]
    elif i==1 and j==2:
      M_inv[i,j]=eff_mass_inv[5]
      M_inv[j,i]=eff_mass_inv[5]

  for n in range(bnd):
    Eaux = np.reshape(np.repeat(arrays['E_k'][:,n,ispin],esize), (snktot,esize))
    delk = (np.reshape(np.repeat(arrays['deltakp'][:,n,ispin],esize), (snktot,esize)) if smearing!=None else None)
    if smearing is None:
      Eaux -= ene
      smearA = 1/(4*temp*(np.cosh(Eaux/(2*temp))**2))
    else:
      if smearing == 'gauss':
        smearA = gaussian(Eaux, ene, delk)
      elif smearing == 'm-p':
        smearA = metpax(Eaux, ene, delk)
    for i in range(3):
      for j in range(3):
        for p in range(3):
          for q in range(3):
            for r  in range(3):
              sig_hall[i,j,p,:,n,ispin] += (int(Eijk(p,q,r))*velkp[:,i,n,ispin]*velkp[:,r,n,ispin]*M_inv[j,q,:,n,ispin])
          L_hall[i,j,p,:] += np.sum(kq_wght*arrays['scattering_tau'][:,n,ispin]**2*sig_hall[i,j,p,:,n,ispin]*(smearA).T, axis=1)
  return L_hall



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from mpi4py import MPI
from .smearing import intmetpax

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def E_Fermi ( Hksp, data_controller, parallel=False ):
  # Calculate the Fermi energy using a braketing algorithm

  arry,attr = data_controller.data_dicts()

  insulator = attr['insulator']
  nktot,nbnd = attr['nkpnts'],attr['bnd']
  nelec,dftSO = attr['nelec'],attr['dftSO']
  nawf,_,snktot,nspin = Hksp.shape
  eig = np.zeros((nawf,snktot,nspin))
  Hksp = Hksp.reshape((nawf,nawf,snktot,nspin), order='C')

  for ispin in range(nspin):
    for ik in range(snktot):
      eig[:,ik,ispin] = np.linalg.eigvalsh(Hksp[:,:,ik,ispin])

  if insulator:
    Efr = np.amax(eig[(nelec-1 if dftSO else nelec//2-1)])
    if parallel:
      Efm = np.zeros((1), dtype=float) if rank==0 else None
      comm.Reduce(Efr, Efm, op=MPI.MAX)
      return comm.bcast(Efm[0] if rank==0 else None)
    else:
      return Efr

  else:
    Elw = 1.0e+8
    Eup = -1.0e+8
    eps = 1.0e-10
    degauss = 0.01

    nmbnd = nbnd-1 if nbnd == nawf else nbnd
    for ispin in range(nspin):
      for kp in range(snktot):
        Elw = min(Elw,eig[0,kp,ispin])
        Eup = max(Elw,eig[nmbnd,kp,ispin])

    Eup = Eup + 2 * degauss
    Elw = Elw - 2 * degauss

    # bisection method
    fac = 1 if dftSO else 2
    sumkup_aux = fac*np.sum(intmetpax(eig[:nbnd,:,:],Eup,degauss))
    sumklw_aux = fac*np.sum(intmetpax(eig[:nbnd,:,:],Elw,degauss))

    if parallel:
      sumkup = np.zeros((1), dtype=float) if rank==0 else None
      sumklw = np.zeros((1), dtype=float) if rank==0 else None
      comm.Reduce(sumkup_aux, sumkup, op=MPI.SUM)
      comm.Reduce(sumklw_aux, sumklw, op=MPI.SUM)
      sumkup = comm.bcast(sumkup[0]/nktot if rank==0 else None)
      sumklw = comm.bcast(sumklw[0]/nktot if rank==0 else None)
    else:
      sumkup = sumkup_aux/nktot
      sumklw = sumklw_aux/nktot

    if (sumkup - nelec) < -eps or (sumklw - nelec) > eps:
      if rank == 0:
        print('Error: cannot bracket Ef')

    maxiter = 100
    for i in range(maxiter):

      Ef = (Eup + Elw)/2
      sumkmid_aux = fac*np.sum(intmetpax(eig[:,:,:],Ef,degauss))
      if parallel:
        sumkmid = np.zeros((1,), dtype=float) if rank==0 else None
        comm.Reduce(sumkmid_aux, sumkmid, op=MPI.SUM)
        sumkmid = comm.bcast(sumkmid[0]/nktot if rank==0 else None)
      else:
        sumkmid = sumkmid_aux/nktot

      if np.abs( sumkmid-nelec ) < eps:
        break
      elif sumkmid-nelec < -eps:
        Elw = Ef
      else:
        Eup = Ef

    return Ef



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

from mendeleev import element
import numpy as np

def write2xsf(data_controller,filename,data=None):
	
	arry,attr = data_controller.data_dicts()
	
	fileobj = open(filename,'w')
	atoms = arry['atoms']
	
	fileobj.write('CRYSTAL\n')
	
	fileobj.write('PRIMVEC\n')
	for i in range(3):
		fileobj.write(' %.14f %.14f %.14f\n' % tuple(arry['a_vectors'][i]*attr['alat']))
		
	fileobj.write('PRIMCOORD\n')
	fileobj.write(str(len(atoms))+' 1\n')         
	for na in range(len(atoms)):
		atom = element(arry['atoms'][na])
		fileobj.write(' %2d' % atom.atomic_number)
		fileobj.write(' %20.14f %20.14f %20.14f' % tuple(arry['tau'][na]))
		fileobj.write('\n')
		
	if data is None:
		fileobj.close()
		return
	
	fileobj.write('BEGIN_BLOCK_DATAGRID_3D\n')
	fileobj.write(' data\n')
	fileobj.write(' BEGIN_DATAGRID_3Dgrid#1\n')
	
	data = np.asarray(data)
	if data.dtype == np.complex128:
		data = np.abs(data)**2
		
	shape = data.shape
	fileobj.write('  %d %d %d\n' % (shape[0]+1, shape[1]+1, shape[2]+1))
	
	origin = np.zeros(3)
	fileobj.write('  %f %f %f\n' % tuple(origin))
	# These are the actual coordinates for the data
	for i in range(3):
		fileobj.write('  %f %f %f\n' %
			tuple(np.array(arry['a_vectors'][i])*attr['alat']))
		
	for k in range(shape[2]+1):
		for j in range(shape[1]+1):
			fileobj.write('   ')
			for i in range(shape[0]+1):
				fileobj.write('%12.8e ' % (data[i%shape[0], j%shape[1], k%shape[2]]))
			fileobj.write('\n')
			
	fileobj.write(' END_DATAGRID_3D\n')
	fileobj.write('END_BLOCK_DATAGRID_3D\n')
	fileobj.close()



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def do_momentum ( data_controller ):
  import numpy as np
  from .perturb_split import perturb_split

  arry,attr = data_controller.data_dicts()

  nktot,_,nawf,nawf,nspin = arry['dHksp'].shape

  arry['pksp'] = np.zeros_like(arry['dHksp'])

  for ispin in range(nspin):
    for ik in range(nktot):
      for l in range(3):
         arry['pksp'][ik,l,:,:,ispin],_ = perturb_split(arry['dHksp'][ik,l,:,:,ispin], 
                                                        arry['dHksp'][ik,l,:,:,ispin], 
                                                        arry['v_k'][ik,:,:,ispin],
                                                        arry['degen'][ispin][ik])



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def Slater_Koster( data_controller, params ):
  # generalized Slater-Koster TB model in the two-center approximation (1st nearest neighbors only)
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()
  # Lattice Vectors
  arry['a_vectors'] = np.array(params['model']['a_vectors'])
  attr['alat'] = 1.0
  
  # Atomic coordinates
  natoms = len(params['model']['atoms'])
  tau = np.zeros((natoms,3),dtype=float)
  for ia in range(natoms):
    tau[ia] = np.array(params['model']['atoms'][str(ia)]['tau'])  
  atoms = []
  shells = []
  for ia in range(natoms):
    atoms.append(params['model']['atoms'][str(ia)]['name'])
    shells.append(params['model']['atoms'][str(ia)])
  arry['tau'] = tau
  arry['atoms'] = atoms
  attr['natoms'] = natoms
  arry['shells'] = shells
  attr['nspin'] = 1 # only unpolarized case for now
  attr['dftSO'] = False # no spin-orbit
  
  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  attr['omega'] = volume
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume 

  # dimensions of the supercell for two-center approximation
  nk1 = nk2 = nk3 = 3
  nkpnts = nk1*nk2*nk3
  attr['nk1'] = nk1
  attr['nk2'] = nk2
  attr['nk3'] = nk3
  attr['nkpnts'] = nkpnts
  
  # on site and hopping parameters
  norbitals = np.zeros(natoms,dtype=int)
  for ia in range(natoms):
    norbitals[ia] = len(params['model']['atoms'][str(ia)]['orbitals']) 
  
  nawf = 0
  for ia in range(natoms):
    nawf += norbitals[ia]
  attr['nawf'] = nawf
  attr['bnd'] = nawf
  attr['nbnds'] = nawf
  attr['shift'] = 0

  # generate all the orbitals positions in the supercell
  sctau = np.zeros((natoms,nk1,nk2,nk3,3),dtype=float)
  for i in range(-1,2):
    for j in range(-1,2):
      for k in range(-1,2):
        for ia in range(natoms):
          sctau[ia,i,k,j,:] = tau[ia] + i*arry['a_vectors'][0] + j*arry['a_vectors'][1] + k*arry['a_vectors'][2]
  sctau = np.reshape(sctau,(natoms*27,3),order='C')
  # make the list of neighbors and find cutoff for two-center approximation
  distance = lambda x,y : np.sqrt(np.sum((x-y)**2))
  cosines = lambda x,y : (y-x)/np.sqrt(np.sum((x-y)**2))
  dist = []
  for ia in range(natoms):
      for n in range(natoms*27):
          dist.append(distance(tau[ia],sctau[n]))
  cutoff = np.sort(np.unique(dist))[1]+(np.sort(np.unique(dist))[2]-np.sort(np.unique(dist))[1])/2
  sctau = np.reshape(sctau,(natoms,nk1,nk2,nk3,3),order='C')
  
  # debug
  arry['sctau'] = sctau
  attr['cutoff'] = cutoff
  arry['norbitals'] = norbitals
  
  HRs = np.zeros((nawf,nawf,nk1,nk2,nk3,1),dtype=complex)

  # on-site matrix elements
  for ia in range(natoms):
    for no in range(norbitals[ia]):
      HRs [ia*norbitals[ia]+no,ia*norbitals[ia]+no,0,0,0,0] = \
      params['model']['atoms'][str(ia)][params['model']['atoms'][str(ia)]['orbitals'][no]]

  # hopping matrix elements
  for i in range(-1,2):
    for j in range(-1,2):
      for k in range(-1,2):
        for ia in range(natoms):
          for ib in range(natoms):
            if distance(tau[ia],sctau[ib,i,j,k,:]) > 0 and distance(tau[ia],sctau[ib,i,j,k,:]) < cutoff:
              lx = cosines(tau[ia],sctau[ib,i,j,k,:])[0]
              ly = cosines(tau[ia],sctau[ib,i,j,k,:])[1]
              lz = cosines(tau[ia],sctau[ib,i,j,k,:])[2]
              
              for noa in range(norbitals[ia]):
                for nob in range(norbitals[ib]):
#                  print(ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k)
                  if noa == 0 and nob == 0:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = params['model']['hoppings']['sss']
                  elif noa == 0 and nob == 1:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = lx*params['model']['hoppings']['sps']
                  elif noa == 0 and nob == 2:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = ly*params['model']['hoppings']['sps']
                  elif noa == 0 and nob == 3:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = lz*params['model']['hoppings']['sps']
                  elif noa == 1 and nob == 1:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx**2*params['model']['hoppings']['pps']+(1.0-lx**2)*params['model']['hoppings']['ppp']
                  elif noa == 2 and nob == 2:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    ly**2*params['model']['hoppings']['pps'] + (1.0-ly**2)*params['model']['hoppings']['ppp']
                  elif noa == 3 and nob == 3:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lz**2*params['model']['hoppings']['pps'] + (1.0-lz**2)*params['model']['hoppings']['ppp']
                  elif noa == 1 and nob == 2:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx*ly*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  elif noa == 2 and nob == 3:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    ly*lz*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  elif noa == 1 and nob == 3:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx*lz*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  elif (noa == 1 and nob == 0): 
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = -lx*params['model']['hoppings']['sps']
                  elif (noa == 2 and nob == 0):
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = -ly*params['model']['hoppings']['sps']
                  elif (noa == 3 and nob == 0):
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = -lz*params['model']['hoppings']['sps']
                  elif noa == 1 and nob == 1:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx**2*params['model']['hoppings']['pps']+(1.0-lx**2)*params['model']['hoppings']['ppp']
                  elif noa == 2 and nob == 2:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    ly**2*params['model']['hoppings']['pps'] + (1.0-ly**2)*params['model']['hoppings']['ppp']
                  elif noa == 3 and nob == 3:
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lz**2*params['model']['hoppings']['pps'] + (1.0-lz**2)*params['model']['hoppings']['ppp']
                  elif (noa == 2 and nob == 1):
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx*ly*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  elif (noa == 3 and nob == 2):
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    ly*lz*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  elif (noa == 3 and nob == 1):
                    HRs[ia*norbitals[ia]+noa,ib*norbitals[ib]+nob,i,j,k,0] = \
                    lx*lz*(params['model']['hoppings']['pps'] - params['model']['hoppings']['ppp'])
                  
    arry['HRs'] = HRs

def graphene( data_controller, params ):
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()

  attr['nk1'] = 3
  attr['nk2'] = 3
  attr['nk3'] = 1

  attr['nawf'] = 2
  attr['nspin'] = 1
  attr['natoms'] = 2

  attr['alat'] = 2.46 * ANGSTROM_AU

  arry['HRs'] = np.zeros((attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin']),dtype=complex)

  # H00
  arry['HRs'][0,1,0,0,0,0] = params['t']
  arry['HRs'][1,0,0,0,0,0] = params['t']

  # H10
  arry['HRs'][1,0,1,0,0,0] = params['t']

  #H20
  arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T

  #H01
  arry['HRs'][1,0,0,1,0,0] = params['t']

  #H02
  arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T

  # Lattice Vectors
  arry['a_vectors'] = np.zeros((3,3),dtype=float)
  arry['a_vectors'] = np.array([[1., 0, 0], [0.5, 3 ** .5 / 2, 0], [0, 0, 10]])
  arry['a_vectors'] = arry['a_vectors']

  # Atomic coordinates
  arry['tau'] = np.zeros((2,3),dtype=float) 

  arry['tau'][0,0] = 0.50000 ;  arry['tau'][0,1] = 0.28867
  arry['tau'][1,0] = 1.00000 ;  arry['tau'][1,1] = 0.57735


  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume 

  arry['atoms']=['C','C']

def graphene2( data_controller, params ):
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()

  attr['nk1'] = 3
  attr['nk2'] = 3
  attr['nk3'] = 1

  attr['nawf'] = 2
  attr['nspin'] = 1
  attr['natoms'] = 2

  arry['naw'] = np.array([1,1])

  attr['alat'] = 2.46 * ANGSTROM_AU

  arry['HRs'] = np.zeros((attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin']),dtype=complex)

  # H00
  arry['HRs'][0,0,0,0,0,0] = params['delta']/2
  arry['HRs'][1,1,0,0,0,0] = -params['delta']/2

  # H00
  arry['HRs'][0,1,0,0,0,0] = params['t']
  arry['HRs'][1,0,0,0,0,0] = params['t']

  # H10
  arry['HRs'][1,0,1,0,0,0] = params['t']

  #H20
  arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T

  #H01
  arry['HRs'][1,0,0,1,0,0] = params['t']

  #H02
  arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T

  # Lattice Vectors
  arry['a_vectors'] = np.zeros((3,3),dtype=float)
  arry['a_vectors'] = np.array([[1., 0, 0], [0.5, 3 ** .5 / 2, 0], [0, 0, 10]])
  arry['a_vectors'] = arry['a_vectors']

  # Atomic coordinates
  arry['tau'] = np.zeros((2,3),dtype=float) 

  arry['tau'][0,0] = 0.50000 ;  arry['tau'][0,1] = 0.28867
  arry['tau'][1,0] = 1.00000 ;  arry['tau'][1,1] = 0.57735


  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume 

  arry['atoms']=['C','C']


def cubium( data_controller, params ):
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()

  attr['nk1'] = 3
  attr['nk2'] = 3
  attr['nk3'] = 3
  attr['Efermi'] = 6*params['t']
  attr['nawf'] = 1
  attr['nspin'] = 1
  attr['natoms'] = 1
  attr['bnd']=1
  attr['shift']=0
  attr['dftSO']=False
  attr['nkpnts']=attr['nk1']*attr['nk2']*attr['nk3']
  attr['nbnds']=1
  attr['nelec']=2

  attr['alat'] = 1.0*ANGSTROM_AU
  attr['omega'] = attr['alat']**3

  arry['HRs'] = np.zeros((attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin']),dtype=complex)

  # H000
  arry['HRs'][0,0,0,0,0,0] = 0.0 - attr['Efermi']

  # H100
  arry['HRs'][0,0,1,0,0,0] = params['t']

  #H200
  arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T

  #H010
  arry['HRs'][0,0,0,1,0,0] = params['t']

  #H020
  arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T

  #H001
  arry['HRs'][0,0,0,0,1,0] = params['t']

  #H002
  arry['HRs'][:,:,0,0,2,0] = np.conj(arry['HRs'][:,:,0,0,1,0]).T

  # Lattice Vectors
  arry['a_vectors'] = np.zeros((3,3),dtype=float)
  arry['a_vectors'] = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

  # Atomic coordinates
  arry['tau'] = np.zeros((1,3),dtype=float) 

  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume 

  arry['atoms']=["Cu"]
  
def cubium2( data_controller, params ):
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()

  attr['nk1'] = 3
  attr['nk2'] = 3
  attr['nk3'] = 3

  attr['nawf'] = 2
  attr['nspin'] = 1
  attr['natoms'] = 1
  attr['bnd']=2
  attr['shift']=0
  attr['dftSO']=False
  attr['nkpnts']=attr['nk1']*attr['nk2']*attr['nk3']
  attr['nbnds']=2
  attr['nelec']=2
  attr['alat'] = 1.0*ANGSTROM_AU
  attr['omega'] = attr['alat']**3

  arry['HRs'] = np.zeros((attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin']),dtype=complex)

  # H000
  arry['HRs'][0,0,0,0,0,0] = -params['Eg']/2 -6.0*params['t']
  arry['HRs'][1,1,0,0,0,0] = params['Eg']/2 +6.0*params['t']

  # H100
  arry['HRs'][0,0,1,0,0,0] = params['t']
  arry['HRs'][1,1,1,0,0,0] = -params['t']

  #H200
  arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T

  #H010
  arry['HRs'][0,0,0,1,0,0] = params['t']
  arry['HRs'][1,1,0,1,0,0] = -params['t']

  #H020
  arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T

  #H001
  arry['HRs'][0,0,0,0,1,0] = params['t']
  arry['HRs'][1,1,0,0,1,0] = -params['t']

  #H002
  arry['HRs'][:,:,0,0,2,0] = np.conj(arry['HRs'][:,:,0,0,1,0]).T

  # Lattice Vectors
  arry['a_vectors'] = np.zeros((3,3),dtype=float)
  arry['a_vectors'] = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

  # Atomic coordinates
  arry['tau'] = np.zeros((1,3),dtype=float) 

  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume 

  arry['atoms']=["Cu"]


def Kane_Mele( data_controller, params ):
  from .constants import ANGSTROM_AU
  from scipy.fftpack import fftshift
  from mpi4py import MPI
  import numpy as np

  arry,attr = data_controller.data_dicts()

  attr['nk1'] = 3
  attr['nk2'] = 3
  attr['nk3'] = 1

  attr['nawf'] = 4
  attr['nspin'] = 1
  attr['natoms'] = 2

  arry['naw'] = [2,2]

  if 'alat' not in params: alat = 1.0
  else: alat = params['alat']

  attr['alat'] = alat * ANGSTROM_AU

  t = params['t']
  soc_par = params['soc_par']
  r_par = params['r_par']
  v_par = params['v_par']

  arry['HRs'] = np.zeros((attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin']),dtype=complex)

  # H00
  arry['HRs'][0,0,0,0,0,0] = t * v_par
  arry['HRs'][1,1,0,0,0,0] = t * v_par
  arry['HRs'][2,2,0,0,0,0] = -t * v_par
  arry['HRs'][3,3,0,0,0,0] = -t * v_par

  # H00
  arry['HRs'][0,2,0,0,0,0] = t
  arry['HRs'][1,3,0,0,0,0] = t
  arry['HRs'][2,0,0,0,0,0] = t
  arry['HRs'][3,1,0,0,0,0] = t

  # H10
  arry['HRs'][2,0,1,0,0,0] = t
  arry['HRs'][3,1,1,0,0,0] = t

  arry['HRs'][0,0,1,0,0,0] = - complex(0.0,soc_par)
  arry['HRs'][1,1,1,0,0,0] =   complex(0.0,soc_par)
  arry['HRs'][2,2,1,0,0,0] =   complex(0.0,soc_par)
  arry['HRs'][3,3,1,0,0,0] = - complex(0.0,soc_par)

  ##H20
  #arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T

  #H01
  arry['HRs'][2,0,0,1,0,0] = t
  arry['HRs'][3,1,0,1,0,0] = t

  arry['HRs'][0,0,0,1,0,0] =   complex(0.0,soc_par)
  arry['HRs'][1,1,0,1,0,0] = - complex(0.0,soc_par)
  arry['HRs'][2,2,0,1,0,0] = - complex(0.0,soc_par)
  arry['HRs'][3,3,0,1,0,0] =   complex(0.0,soc_par)

  ##H02
  #arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T

  #H21
  arry['HRs'][0,0,2,1,0,0] = - complex(0.0,soc_par)
  arry['HRs'][1,1,2,1,0,0] =   complex(0.0,soc_par)
  arry['HRs'][2,2,2,1,0,0] =   complex(0.0,soc_par)
  arry['HRs'][3,3,2,1,0,0] = - complex(0.0,soc_par)

  ##H12
  ##arry['HRs'][:,:,1,2,0,0] = np.conj(arry['HRs'][:,:,2,1,0,0]).T

  r3h =np.sqrt(3.0)/2.0

  arry['HRs'][0,3,0,0,0,0] += r_par * complex(-r3h,0.5) # 1j * r_par * (0.5 * 1 - r3h * -1j) 
  arry['HRs'][1,2,0,0,0,0] += r_par * complex(r3h,0.5) #1j * r_par * (0.5 * 1 - r3h * 1j)
  arry['HRs'][3,0,0,0,0,0] += r_par * complex(-r3h,-0.5)
  arry['HRs'][2,1,0,0,0,0] += r_par * complex(r3h,-0.5)

  arry['HRs'][0,3,1,0,0,0] += -r_par * complex(r3h,0.5) # -1j * r_par * (0.5 * 1 + r3h * -1j)
  arry['HRs'][1,2,1,0,0,0] += -r_par * complex(-r3h,0.5) # -1j * r_par * (0.5 * 1 + r3h * 1j)

  arry['HRs'][0,3,0,1,0,0] += complex(0.0,r_par) # -1j * r_par * -1 * 1
  arry['HRs'][1,2,0,1,0,0] += complex(0.0,r_par) # -1j * r_par * -1 * 1

  #H02
  arry['HRs'][:,:,0,2,0,0] = np.conj(arry['HRs'][:,:,0,1,0,0]).T
  #H20
  arry['HRs'][:,:,2,0,0,0] = np.conj(arry['HRs'][:,:,1,0,0,0]).T
  #H12
  arry['HRs'][:,:,1,2,0,0] = np.conj(arry['HRs'][:,:,2,1,0,0]).T

  # Lattice Vectors
  arry['a_vectors'] = np.zeros((3,3),dtype=float)
  arry['a_vectors'] = np.array([[1., 0, 0], [0.5, 3 ** .5 / 2, 0], [0, 0, 10]])
  arry['a_vectors'] = arry['a_vectors']

  # Atomic coordinates
  arry['tau'] = np.zeros((2,3),dtype=float)
  arry['tau'][0] = np.dot([1/3,1/3,0.],arry['a_vectors'])
  arry['tau'][1] = np.dot([2/3,2/3,0.],arry['a_vectors'])

  # Reciprocal Lattice
  arry['b_vectors'] = np.zeros((3,3),dtype=float)
  volume = np.dot(np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]),arry['a_vectors'][2,:])
  arry['b_vectors'][0,:] = (np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))/volume
  arry['b_vectors'][1,:] = (np.cross(arry['a_vectors'][2,:],arry['a_vectors'][0,:]))/volume
  arry['b_vectors'][2,:] = (np.cross(arry['a_vectors'][0,:],arry['a_vectors'][1,:]))/volume

  attr['omega'] = alat**3 * arry['a_vectors'][0,:].dot(np.cross(arry['a_vectors'][1,:],arry['a_vectors'][2,:]))

  arry['species']=["KM","KM"]


def build_TB_model ( data_controller, parameters ):
  if parameters['label'].upper() == 'GRAPHENE':
    graphene(data_controller, parameters)
  elif parameters['label'].upper() == 'GRAPHENE2':
    graphene2(data_controller, parameters)
  elif parameters['label'].upper() == 'CUBIUM':
    cubium(data_controller, parameters)
  elif parameters['label'].upper() == 'CUBIUM2':
    cubium2(data_controller, parameters)
  elif parameters['label'].upper() == 'KANE_MELE':
    Kane_Mele(data_controller, parameters)
  elif parameters['label'].upper() == 'SLATER_KOSTER':
    Slater_Koster(data_controller, parameters)
  else:
    print('ERROR: Label "%s" not found in builtin models.'%label)




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

def inverse_participation_ratio (data_controller):

  arry,attr = data_controller.data_dicts()

  nbands = attr['bnd']

  if 'Hksp' in arry:
    kpts = arry['kpnts']
  else:
    kpts = arry['kq'].T

  nkpts = kpts.shape[0]

  nspin = attr['nspin']

  ipr = np.zeros((nspin,nkpts,nbands,3),dtype=object)

  for ispin in range(nspin):
    for ikpt in range(nkpts):
      for iband in range(nbands):

        vk = arry['v_k'][ikpt,:,iband,ispin]
        ek = arry['E_k'][ikpt,iband,ispin]

        vk_abs = np.abs(vk)

        ipr[ispin,ikpt,iband,0] = kpts[ikpt]
        ipr[ispin,ikpt,iband,1] = ek
        ipr[ispin,ikpt,iband,2] = np.sum(vk_abs**4) / ( np.sum(vk_abs**2)**2 )

  return ipr




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

class TauModel:
  '''
  Object containing a functional form for Tau, depending on termperature and an energy eigenvalue (both in units of eV).
  All TauModel's included in the scattering_channels arugment of PAOFLOW's transport routine
   are homically summed to obtain the effective relaxation time for each (temp,eig) combination.
  An additional argument, params (a dictionary), is required. Further parameters can be passed into    the routine by including them in this dictionary.
  '''

  def __init__ ( self, function=None, params=None, weight=1. ):
    '''
    Arguments:
      function (func): A function requiring 3 arguments, (temp,eig,params)
      params (dict): A dictionary with any additional constants of variables the function may require
      weight (float): A weight, w_i, incorporated in the harmonic sum of tau. 1/Tau = Sum(w_i/Tau_i)
    '''
    self.function = function
    self.params = params
    self.weight = weight

  def evaluate ( self, temp, eigs ):
    return self.function(temp, eigs, self.params)



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

class CGBF:
  '''Class for a contracted Gaussian basis function'''
  def __init__ ( self, origin, atid=0 ):
    self.origin = tuple(float(i) for i in origin)
    self.powers = [] 
    self.pnorms = []
    self.prims = []
    self.pnorms = []
    self.pexps = []
    self.pcoefs = []
    self.atid = atid


def gammln ( x ):
  "Numerical recipes, section 6.1"
  cof = [76.18009172947146, -86.50532032941677,
         24.01409824083091, -1.231739572450155,
         0.1208650973866179e-2, -0.5395239384953e-5]
  xt = x
  tmp = x + 5.5
  tmp = tmp - (x+0.5)*np.log(tmp)
  ser = 1.000000000190015
  for c in cof:
    xt += 1
    ser += c / xt
  return -tmp + np.log(2.5066282746310005*ser/x);


def _gser ( a, x ):
  "Series representation of Gamma. NumRec sect 6.1."
  ITMAX = 100
  EPS = 3.e-7

  gln = gammln(a)
  if x < 0:
    raise ValueError(f'Bounds: x >= 0, x={x}')

  if x == 0:
    return 0, gln

  ap = a
  delt = tsum = 1/a
  for i in range(ITMAX):
    ap = 1 + ap
    delt *= x / ap
    tsum += delt
    if abs(delt) < abs(tsum)*EPS:
      break
  else:
    print('a too large, ITMAX too small in gser')

  gamser = tsum * np.exp(-x+a*np.log(x)-gln)
  return gamser, gln


def _gcf ( a, x ):
  "Continued fraction representation of Gamma. NumRec sect 6.1"
  ITMAX = 100
  EPS = 3.e-7
  FPMIN = 1.e-30

  gln = gammln(a)
  b = 1 + x - a
  c = 1 / FPMIN
  d = 1 / b
  h = d
  for i in range(1, ITMAX+1):
    an = -i*(i-a)
    b = b + 2
    d = an*d + b
    if abs(d) < FPMIN:
      d = FPMIN
    c = b + an/c
    if abs(c) < FPMIN:
      c = FPMIN
    d = 1 / d
    delt = d * c
    h *= delt
    if abs(delt-1) < EPS:
      break
  else:
    print('a too large, ITMAX too small in gcf')

  gammcf = h * np.exp(-x+a*np.log(x)-gln)
  return gammcf, gln


def dist ( A, B ):
  return np.sum([(A[i]-B[i])**2 for i in range(3)])


def gaussian_product_center ( alpha1, A, alpha2, B ):
  gamma = alpha1 + alpha2
  return [(alpha1*A[i]+alpha2*B[i])/gamma for i in range(3)]


def Fgamma ( m, x ):
  "Incomplete gamma function"
  mp5 = m + 0.5
  x = max(abs(x), 1e-8)
  return gamm_inc(mp5, x) / (2*x**mp5)


def gammp ( a, x ):
  "Returns the incomplete gamma function P(a;x). NumRec sect 6.2."
  if x <= 0 or a < 0:
    raise ValueError(f'Bounds: (x > 0) and (a >= 0), x={x}, a={a}')

  if x < a + 1: 
    return _gser(a, x)

  gammcf,gln = _gcf(a, x)
  return 1-gammcf, gln 


def gamm_inc ( a, x ):
  "Incomple gamma function; computed from NumRec routine gammp."
  gammap,gln = gammp(a, x)
  return gammap * np.exp(gln)


def fact ( i ):
  "Normal factorial"
  val = 1
  while i > 1:
    val *= i
    i -= 1
  return val


def fact_ratio ( a, b ):
  return fact(a) / fact(b) / fact(a-2*b)


def binomial ( a, b ):
  "Binomial coefficient"
  return fact(a) / fact(b) / fact(a-b)


def binomial_prefactor ( s, ia, ib, xpa, xpb ):
  "From Augspurger and Dykstra"
  bsum = 0
  for t in range(s+1):
    if s-ia <= t <= ib:
      sgn = (xpa)**(ia-s+t) * (xpb)**(ib-t)
      bsum += sgn * binomial(ia, s-t) * binomial(ib, t)
  return bsum


def B0 ( i, r, g ):
  return fact_ratio(i,r) * (4*g)**(r-i)


def fB ( i, l1, l2, P, A, B, r, g ):
  return binomial_prefactor(i,l1,l2,P-A,P-B) * B0(i,r,g)


def B_array ( l1, l2, l3, l4, p, a, b, q, c, d, g1, g2, delta ):
  ind_max = l1 + l2 + l3 + l4 + 1
  B = np.zeros(ind_max, dtype=float)
  for i1 in range(l1+l2+1):
    for i2 in range(l3+l4+1):
      for r1 in range(i1//2+1):
        for r2 in range(i2//2+1):
          for u in range((i1+i2)//2-r1-r2+1):
            "THO eq. 2.22"
            dm = i1 + i2 - 2*(r1+r2)
            ind = dm - u
            fr = fact_ratio(dm, u)
            pqp = (q-p)**(dm-2*u)
            pdelta = delta**ind
            sign = (-1)**i2 * (-1)**u
            fb1 = fB(i1, l1, l2, p, a, b, r1, g1)
            fb2 = fB(i2, l3, l4, q, c, d, r2, g2)
            B[ind] += sign * fb1 * fb2 * fr * pqp / pdelta
  return B


def coulomb_repulsion(axyz, anorm, almn, aalpha,
                      bxyz, bnorm, blmn, balpha,
                      cxyz, cnorm, clmn, calpha,
                      dxyz, dnorm, dlmn, dalpha ):

  rab = dist(axyz, bxyz)
  rcd = dist(cxyz, dxyz)
  pxyz = gaussian_product_center(aalpha, axyz, balpha, bxyz)
  qxyz = gaussian_product_center(calpha, cxyz, dalpha, dxyz)
  rpq = dist(pxyz, qxyz)
  gamma1 = aalpha + balpha
  gamma2 = calpha + dalpha
  delta = (1/gamma1 + 1/gamma2) / 4

  Bx,By,Bz = (B_array(almn[i], blmn[i], clmn[i], dlmn[i],
               pxyz[i], axyz[i], bxyz[i], qxyz[i], cxyz[i], dxyz[i],
               gamma1, gamma2, delta) for i in range(3))

  i = 0
  bsum = 0
  nl = almn[0]
  for i in range(almn[0]+blmn[0]+clmn[0]+dlmn[0]+1):
    for j in range(almn[1]+blmn[1]+clmn[1]+dlmn[1]+1):
      for k in range(almn[2]+blmn[2]+clmn[2]+dlmn[2]+1):
        bsum += Bx[i]*By[j]*Bz[k]*Fgamma(i+j+k, rpq/delta/4)

  norm = anorm * bnorm * cnorm * dnorm
  g2 = gamma1 * gamma2 * np.sqrt(gamma1+gamma2)
  e1 = np.exp(-aalpha*balpha*rab/gamma1)
  e2 = np.exp(-calpha*dalpha*rcd/gamma2)
  return 2 * np.pi**2.5 * e1 * e2 * bsum * norm / g2


def contr_coulomb ( aexps, acoefs, anorms, xyza, powa,
                    bexps, bcoefs, bnorms, xyzb, powb,
                    cexps, ccoefs, cnorms, xyzc, powc,
                    dexps, dcoefs, dnorms, xyzd, powd ):

  Jij = 0.
  for i in range(len(aexps)):
    for j in range(len(bexps)):
      for k in range(len(cexps)):
        for l in range(len(dexps)):
          incr = coulomb_repulsion(xyza, anorms[i], powa[i], aexps[i],
                                   xyzb, bnorms[j], powb[j], bexps[j],
                                   xyzc, cnorms[k], powc[k], cexps[k],
                                   xyzd, dnorms[l], powd[l], dexps[l])
          Jij += incr * acoefs[i] * bcoefs[j] * ccoefs[k] * dcoefs[l]
  return Jij



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def do_spin_texture ( data_controller ):
  import os
  import numpy as np
  from mpi4py import MPI
  from .communication import gather_full

  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  arrays = data_controller.data_arrays
  attributes = data_controller.data_attributes

  fermi_up,fermi_dw = attributes['fermi_up'],attributes['fermi_dw']
  nawf,nk1,nk2,nk3 = attributes['nawf'],attributes['nk1'],attributes['nk2'],attributes['nk3']
  E_k_full = gather_full(arrays['E_k'], attributes['npool'])
 
  ind_plot = []
  icount = None
  if rank == 0:
    icount = 0
    for ib in range(nawf):
      E_k_min = np.amin(E_k_full[:,ib,0])
      E_k_max = np.amax(E_k_full[:,ib,0])
      btwUp = (E_k_min < fermi_up and E_k_max > fermi_up)
      btwDwn = (E_k_min < fermi_dw and E_k_max > fermi_dw)
      btwUaD = (E_k_min > fermi_dw and E_k_max < fermi_up)
      if btwUp or btwDwn or btwUaD:
        ind_plot.append(ib)
        icount += 1

  icount = comm.bcast(icount)
  ind_plot = comm.bcast(ind_plot)

  Sj = arrays['Sj']
  snktot = arrays['v_k'].shape[0]
  sktxtaux = np.zeros((snktot,3,nawf,nawf), dtype=complex)

  # Compute matrix elements of the spin operator
  for ik in range(snktot):
      for l in range(3):
        sktxtaux[ik,l,:,:] = np.conj(arrays['v_k'][ik,:,:,0].T).dot(Sj[l,:,:]).dot(arrays['v_k'][ik,:,:,0])

  sktxtaux = np.take(np.diagonal(sktxtaux,axis1=2,axis2=3), ind_plot, axis=2)
  sktxt = gather_full(np.ascontiguousarray(sktxtaux), attributes['npool'])
  sktxtaux = None

  if rank == 0:
    if 'kq' in arrays and E_k_full.shape[0] == arrays['kq'].shape[1]:
      f=open(os.path.join(attributes['opath'],'spin-texture-bands'+'.dat'),'w')
      for ik in range(E_k_full.shape[0]):
        for ib in range(icount):
          idx=ind_plot[ib]
          f.write('\t'.join(['%d'%ik]+['% 5.8f'%E_k_full[ik,idx]]+['% 5.8f'%j for j in sktxt[ik,:,ib].real])+'\n')
        f.write("\n")
      f.close()
    else:
      sktxt = np.reshape(sktxt, (nk1,nk2,nk3,3,icount), order='C')
      for ib in range(icount):
        np.savez(os.path.join(attributes['opath'],'spin_text_band_'+str(ib)), spinband=sktxt[:,:,:,:,ib])


  sktxt = None
  E_k_full = None
